"""Test cases for the srf_filter module."""

import datetime

import numpy as np
import pytest
import xarray as xr

from eradiate.srf_tools import integral_filter, spectral_filter, threshold_filter, trim
from eradiate.units import unit_registry as ureg

SRF_VALUES_RELEVANT = np.array([0.0, 0.5, 1.0, 0.5, 0.0])


@pytest.fixture
def srf_values_trailing_and_leading_zeros() -> np.ndarray:
    return np.concatenate(
        [
            np.zeros(50),
            SRF_VALUES_RELEVANT,
            np.zeros(200),
        ]
    )


@pytest.fixture
def srf_values_trailing_zeros() -> np.ndarray:
    return np.concatenate(
        [
            SRF_VALUES_RELEVANT,
            np.zeros(200),
        ]
    )


@pytest.fixture
def srf_values_leading_zeros() -> np.ndarray:
    return np.concatenate(
        [
            np.zeros(50),
            SRF_VALUES_RELEVANT,
        ]
    )


@pytest.mark.parametrize(
    "srf_values",
    [
        "srf_values_trailing_and_leading_zeros",
        "srf_values_trailing_zeros",
        "srf_values_leading_zeros",
    ],
)
def test_trim(request, srf_values) -> None:
    """Trim all leading zeros except last and all trailing zeros except first."""
    srf_values = request.getfixturevalue(srf_values)
    utcnow = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
    ds = xr.Dataset(
        {"srf": ("w", srf_values, {"units": "dimensionless"})},
        coords={"w": ("w", np.linspace(300, 800, srf_values.size), {"units": "nm"})},
        attrs={"history": f"{utcnow} - data set creation"},
    )
    trimmed = trim(srf=ds)
    assert trimmed.srf.size == SRF_VALUES_RELEVANT.size


def test_threshold_filter() -> None:
    """
    Drop data points where response is smaller or equal than a threshold value.
    """
    srf_values = np.array([1e-6, 1e-3, 1e-2, 1e0, 1e-1, 1e-2, 1e-2, 1e-5])
    utcnow = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
    srf = xr.Dataset(
        {"srf": ("w", srf_values, {"units": "dimensionless"})},
        coords={"w": ("w", np.linspace(300, 800, srf_values.size), {"units": "nm"})},
        attrs={"history": f"{utcnow} - data set creation"},
    )
    threshold = 0.002
    filtered = threshold_filter(srf=srf, value=threshold)

    assert srf_values[srf_values > threshold].size == filtered.srf.size


@pytest.mark.parametrize(
    "wrange",
    [
        (500.0 * ureg.nm, None),
        (None, 600.0 * ureg.nm),
        (500.0 * ureg.nm, 600.0 * ureg.nm),
        (None, None),
    ],
)
def test_spectral_filter(wrange) -> None:
    """
    Drop points falling out of wavelength range specified by wmin and wmax.
    """
    wmin, wmax = wrange

    w = np.linspace(400, 800)
    srf_values = np.random.random(w.size)
    utcnow = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
    srf = xr.Dataset(
        {"srf": ("w", srf_values, {"units": "dimensionless"})},
        coords={"w": ("w", w, {"units": "nm"})},
        attrs={"history": f"{utcnow} - data set creation"},
    )
    filtered = spectral_filter(srf=srf, wmin=wmin, wmax=wmax)

    if wmax is None and wmin is not None:
        assert srf_values[w >= wmin.m].size == filtered.srf.size
    if wmin is None and wmax is not None:
        assert srf_values[w <= wmax.m].size == filtered.srf.size
    if wmin is not None and wmax is not None:
        assert srf_values[(w <= wmax.m) & (w >= wmin.m)].size == filtered.srf.size
    if wmin is None and wmax is None:
        assert srf_values.size == filtered.srf.size


@pytest.mark.parametrize(
    "percentage",
    [50.0, 90.0, 95.0, 99.0],
)
def test_integral_filter(percentage) -> None:
    """
    Keep only data that contribute to the integrated spectral response value
    to the amount of the specified percentage.
    """

    srf_values = np.ones(100000)
    w_values = np.linspace(300, 800, srf_values.size)

    utcnow = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
    srf = xr.Dataset(
        {"srf": ("w", srf_values, {"units": "dimensionless"})},
        coords={"w": ("w", w_values, {"units": "nm"})},
        attrs={"history": f"{utcnow} - data set creation"},
    )
    filtered = integral_filter(srf=srf, percentage=percentage)

    assert int(100 * filtered.srf.size / srf_values.size) == percentage



import drjit as dr
import mitsuba as mi
import numpy as np
import pytest

from eradiate import unit_context_config as ucc
from eradiate import unit_context_kernel as uck
from eradiate import unit_registry as ureg
from eradiate.scenes.measure import Target, TargetPoint, TargetRectangle


def test_target_origin(modes_all):
    # TargetPoint: basic constructor
    with ucc.override({"length": "km"}):
        t = TargetPoint([0, 0, 0])
        assert t.xyz.units == ureg.km

    with pytest.raises(ValueError):
        TargetPoint(0)

    # TargetPoint: check kernel item
    with ucc.override({"length": "km"}), uck.override({"length": "m"}):
        t = TargetPoint([1, 2, 0])
        assert dr.allclose(t.kernel_item(), [1000, 2000, 0])

    # TargetRectangle: basic constructor
    with ucc.override({"length": "km"}):
        t = TargetRectangle(0, 1, 0, 1)
        assert t.xmin == 0.0 * ureg.km
        assert t.xmax == 1.0 * ureg.km
        assert t.ymin == 0.0 * ureg.km
        assert t.ymax == 1.0 * ureg.km

    with ucc.override({"length": "m"}):
        t = TargetRectangle(0, 1, 0, 1)
        assert t.xmin == 0.0 * ureg.m
        assert t.xmax == 1.0 * ureg.m
        assert t.ymin == 0.0 * ureg.m
        assert t.ymax == 1.0 * ureg.m

    with pytest.raises(ValueError):
        TargetRectangle(0, 1, "a", 1)

    with pytest.raises(ValueError):
        TargetRectangle(0, 1, 1, -1)

    # TargetRectangle: check kernel item
    t = TargetRectangle(-1, 1, -1, 1)

    with uck.override({"length": "mm"}):  # Tricky: we can't compare transforms directly
        kernel_item = t.kernel_item()["to_world"]
        assert dr.allclose(
            kernel_item.transform_affine(mi.Point3f(-1, -1, 0)), [-1000, -1000, 0]
        )
        assert dr.allclose(
            kernel_item.transform_affine(mi.Point3f(1, 1, 0)), [1000, 1000, 0]
        )
        assert dr.allclose(
            kernel_item.transform_affine(mi.Point3f(1, 1, 42)), [1000, 1000, 42]
        )

    # Factory: basic test
    with ucc.override({"length": "m"}):
        t = Target.new("point", xyz=[1, 1, 0])
        assert isinstance(t, TargetPoint)
        assert np.allclose(t.xyz, ureg.Quantity([1, 1, 0], ureg.m))

        t = Target.new("rectangle", 0, 1, 0, 1)
        assert isinstance(t, TargetRectangle)

    # Converter: basic test
    with ucc.override({"length": "m"}):
        t = Target.convert({"type": "point", "xyz": [1, 1, 0]})
        assert isinstance(t, TargetPoint)
        assert np.allclose(t.xyz, ureg.Quantity([1, 1, 0], ureg.m))

        t = Target.convert([1, 1, 0])
        assert isinstance(t, TargetPoint)
        assert np.allclose(t.xyz, ureg.Quantity([1, 1, 0], ureg.m))

        with pytest.raises(ValueError):
            Target.convert({"xyz": [1, 1, 0]})



import numpy as np
import pytest

from eradiate import unit_registry as ureg
from eradiate.spectral.index import (
    CKDSpectralIndex,
    MonoSpectralIndex,
    SpectralIndex,
)


def test_spectral_index_new(modes_all):
    w = 1.0 * ureg.nm
    si = SpectralIndex.new(w=w)
    assert si.w == w


def test_spectral_index_new_mono():
    w = 1.0 * ureg.nm
    mono = SpectralIndex.new(mode="mono", w=w)
    assert isinstance(mono, MonoSpectralIndex)
    assert mono.w == w
    assert mono.formatted_repr == f"{w:g~P}"


def test_spectral_index_new_ckd():
    w = 1.0 * ureg.nm
    g = 0.5
    ckd = SpectralIndex.new(mode="ckd", w=w, g=g)
    assert isinstance(ckd, CKDSpectralIndex)
    assert ckd.w == w
    assert ckd.g == g
    assert ckd.formatted_repr == f"{w:g~P}:{g:g}"


def test_spectral_index_new_invalid():
    """``SpectralIndex.new()`` raises when arguments are invalid."""

    # mode = "invalid"

    with pytest.raises(KeyError):
        mode_invalid = "invalid"
        SpectralIndex.new(mode=mode_invalid)

    # mode = "mono"

    with pytest.raises(TypeError):
        w_invalid = np.array([1.0, 2.0]) * ureg.nm  # w must be scalar quantity
        SpectralIndex.new(mode="mono", w=w_invalid)

    with pytest.raises(ValueError):
        w_invalid = -1.0 * ureg.nm  # w must be positive
        SpectralIndex.new(mode="mono", w=w_invalid)

    # mode = "ckd"

    with pytest.raises(TypeError):
        w_invalid = np.array([1.0, 2.0]) * ureg.nm  # w must be scalar quantity
        SpectralIndex.new(mode="ckd", w=w_invalid)

    with pytest.raises(ValueError):
        w_invalid = -1.0 * ureg.nm  # w must be positive
        SpectralIndex.new(mode="ckd", w=w_invalid)

    with pytest.raises(ValueError):
        g_invalid = 2.0  # g must be between 0 and 1
        SpectralIndex.new(mode="ckd", g=g_invalid)

    with pytest.raises(ValueError):
        g_invalid = -1.0  # g must be between 0 and 1
        SpectralIndex.new(mode="ckd", g=g_invalid)


def test_spectral_index_from_dict_mono():
    w = 1.0 * ureg.nm
    mono = SpectralIndex.from_dict({"mode": "mono", "w": w})
    assert isinstance(mono, MonoSpectralIndex)
    assert mono.w == w
    assert mono.formatted_repr == f"{w:g~P}"


def test_spectral_index_from_dict_ckd():
    w = 1.0 * ureg.nm
    g = 0.5
    ckd = SpectralIndex.from_dict({"mode": "ckd", "w": w, "g": g})
    assert isinstance(ckd, CKDSpectralIndex)
    assert ckd.w == w
    assert ckd.g == g
    assert ckd.formatted_repr == f"{w:g~P}:{g:g}"


def test_spectral_index_convert_mono():
    w = 1.0 * ureg.nm
    value = {"mode": "mono", "w": w}
    converted = SpectralIndex.convert(value)
    assert isinstance(converted, MonoSpectralIndex)
    assert np.allclose(converted.w, w)


def test_spectral_index_convert_ckd():
    w = 1.0 * ureg.nm
    g = 0.5
    value = {"mode": "ckd", "w": w, "g": g}
    converted = SpectralIndex.convert(value)
    assert isinstance(converted, CKDSpectralIndex)
    assert np.allclose(converted.w, w)
    assert np.allclose(converted.g, g)



import mitsuba as mi
import numpy as np
import pytest

import eradiate
from eradiate import unit_context_config as ucc
from eradiate import unit_registry as ureg
from eradiate.contexts import KernelContext
from eradiate.radprops import ZGrid
from eradiate.scenes.atmosphere import (
    HeterogeneousAtmosphere,
    ParticleLayer,
)
from eradiate.scenes.core import traverse
from eradiate.scenes.geometry import SceneGeometry
from eradiate.spectral.index import CKDSpectralIndex, SpectralIndex
from eradiate.test_tools.types import check_scene_element


def test_heterogeneous_empty(modes_all_double):
    # Passing no component is not allowed
    with pytest.raises(ValueError):
        HeterogeneousAtmosphere()


@pytest.mark.parametrize("geometry", ["plane_parallel", "spherical_shell"])
@pytest.mark.parametrize("component", ["molecular", "particle"])
def test_heterogeneous_single_mono(
    mode_mono, geometry, component, atmosphere_us_standard_mono
):
    """
    Unit tests for a HeterogeneousAtmosphere with a single component.
    """
    # Construct succeeds
    if component == "molecular":
        atmosphere = HeterogeneousAtmosphere(
            geometry=geometry,
            molecular_atmosphere=atmosphere_us_standard_mono,
        )
        si = atmosphere.spectral_set().spectral_indices().__next__()
        kernel_context = KernelContext(si=si)

    else:
        component = ParticleLayer()
        atmosphere = HeterogeneousAtmosphere(
            geometry=geometry, particle_layers=[component]
        )

        kernel_context = KernelContext()

    # The scene element produces valid kernel dictionary specifications
    check_scene_element(atmosphere, ctx=kernel_context)


@pytest.mark.parametrize("geometry", ["plane_parallel", "spherical_shell"])
@pytest.mark.parametrize("component", ["molecular", "particle"])
def test_heterogeneous_single_ckd(
    mode_ckd, geometry, component, atmosphere_us_standard_ckd
):
    """
    Unit tests for a HeterogeneousAtmosphere with a single component.
    """
    # Construct succeeds
    if component == "molecular":
        atmosphere = HeterogeneousAtmosphere(
            geometry=geometry, molecular_atmosphere=atmosphere_us_standard_ckd
        )

    else:
        component = ParticleLayer()
        atmosphere = HeterogeneousAtmosphere(
            geometry=geometry, particle_layers=[component]
        )

    # The scene element produces valid kernel dictionary specifications
    spectral_set = atmosphere.spectral_set()
    if spectral_set is not None:
        si = list(spectral_set.spectral_indices())[0]
    else:
        si = SpectralIndex.new()
    kernel_context = KernelContext(si=si)
    check_scene_element(atmosphere, ctx=kernel_context)


@pytest.mark.slow
@pytest.mark.parametrize("geometry", ["plane_parallel", "spherical_shell"])
def test_heterogeneous_multi_mono(mode_mono, geometry, atmosphere_us_standard_mono):
    """
    Unit tests for a HeterogeneousAtmosphere with multiple (2+) components.
    """
    # Construct succeeds
    atmosphere = HeterogeneousAtmosphere(
        geometry=geometry,
        molecular_atmosphere=atmosphere_us_standard_mono,
        particle_layers=[ParticleLayer() for _ in range(2)],
    )

    # particles phae function is tabulated for wavelengths starting at 350 nm
    for si in atmosphere.spectral_set().spectral_indices():
        if si.w.m > 350:
            break
    kernel_context = KernelContext(si=si)

    # The scene element produces valid kernel dictionary specifications
    check_scene_element(atmosphere, ctx=kernel_context)


@pytest.mark.parametrize("geometry", ["plane_parallel", "spherical_shell"])
def test_heterogeneous_multi_ckd(mode_ckd, geometry, atmosphere_us_standard_ckd):
    """
    Unit tests for a HeterogeneousAtmosphere with multiple (2+) components.
    """
    # Construct succeeds
    atmosphere = HeterogeneousAtmosphere(
        geometry={"type": geometry, "zgrid": np.linspace(0, 120, 121) * ureg.km},
        molecular_atmosphere=atmosphere_us_standard_ckd,
        particle_layers=[ParticleLayer() for _ in range(2)],
    )

    si = CKDSpectralIndex(w=550.0 * ureg.nm, g=0.5)
    kernel_context = KernelContext(si=si)

    # The scene element produces valid kernel dictionary specifications
    check_scene_element(atmosphere, ctx=kernel_context)


@pytest.mark.parametrize("field", ["sigma_a", "sigma_t"])
def test_heterogeneous_mix_collision_coefficients(modes_all_double, field):
    """
    Check for component mixing correctness. We expect that the absorption and
    extinction coefficients properly add up.
    """
    with ucc.override(length="km"):
        component_1 = ParticleLayer(bottom=0.0, top=1.25)
        component_2 = ParticleLayer(bottom=0.5, top=1.5)
        component_3 = ParticleLayer(bottom=0.75, top=2.0)

    mixed = HeterogeneousAtmosphere(
        geometry={
            "type": "plane_parallel",
            "zgrid": np.linspace(0, 120, 1201) * ureg.km,
        },
        particle_layers=[component_1, component_2, component_3],
    )
    ctx = KernelContext()
    zgrid = mixed.geometry.zgrid

    # Evaluate all profiles on the container's altitude grid
    radprofiles = {}

    for component, atmosphere in [
        ("component_1", component_1),
        ("component_2", component_2),
        ("component_3", component_3),
        ("mixed", mixed),
    ]:
        radprofiles[component] = atmosphere.eval_radprops(
            ctx.si, zgrid, optional_fields=True
        )

    collision_coefficient = {}
    for z in [0.1, 0.6, 1.0, 1.4, 1.9] * ureg.km:
        values = {}

        for component, radprofile in radprofiles.items():
            z_units = ureg.Unit(radprofile.coords["z_layer"].attrs["units"])

            field_units = ureg(radprofile.data_vars["sigma_a"].attrs["units"])
            values[component] = (
                float(
                    radprofile.data_vars[field].interp(
                        z_layer=z.m_as(z_units),
                        kwargs={"fill_value": 0.0},
                        method="nearest",
                    )
                )
                * field_units
            )

        collision_coefficient[z.m] = values

    components = sorted(set(radprofiles.keys()) - {"mixed"})

    for z in collision_coefficient.keys():
        total = collision_coefficient[z]["mixed"]
        expected = sum(collision_coefficient[z][component] for component in components)
        np.testing.assert_allclose(
            total.m_as(ureg.m**-1),
            expected.m_as(ureg.m**-1),
            err_msg=f"Failed for altitude {z = }",
        )


def test_heterogeneous_mix_weights(
    modes_all_double, atmosphere_us_standard_mono, atmosphere_us_standard_ckd
):
    """
    Check that component weights are correctly computed.
    """
    ctx = KernelContext()
    geometry = SceneGeometry.convert(
        {
            "type": "plane_parallel",
            "ground_altitude": 0.0 * ureg.km,
            "toa_altitude": 100.0 * ureg.km,
            "zgrid": ZGrid(np.linspace(0, 100, 101) * ureg.km),
        }
    )

    # Fist basic check: a uniform layer and a molecular atmosphere
    mixed = HeterogeneousAtmosphere(
        geometry=geometry,
        molecular_atmosphere=(
            atmosphere_us_standard_ckd
            if eradiate.mode().is_ckd
            else atmosphere_us_standard_mono
        ),
        particle_layers=ParticleLayer(
            bottom=0.0 * ureg.km,
            top=50.0 * ureg.km,
            distribution={"type": "uniform"},
        ),
    )
    template, _ = traverse(mixed.phase)
    mi_phase = mi.load_dict(template.render(ctx))
    mi_params = mi.traverse(mi_phase)

    # Weights should be non-zero over the first 50 km, and 0 above
    # (all to the molecular component)
    weights = np.squeeze(mi_params["weight.data"])
    assert len(weights) == geometry.zgrid.n_layers

    middle = np.argwhere(geometry.zgrid.layers <= 50.0 * ureg.km).max() + 1

    assert np.all((weights[:middle] > 0.0) & (weights[:middle] < 1.0))
    assert np.all(weights[middle:] == 0.0)

    # Second check: simple disjoint components, more than 1
    mixed = HeterogeneousAtmosphere(
        geometry=geometry,
        particle_layers=[
            ParticleLayer(
                bottom=0.0 * ureg.km,
                top=50.0 * ureg.km,
                distribution={"type": "uniform"},
            ),
            ParticleLayer(
                bottom=50.0 * ureg.km,
                top=80.0 * ureg.km,
                distribution={"type": "uniform"},
            ),
            ParticleLayer(
                bottom=80.0 * ureg.km,
                top=100.0 * ureg.km,
                distribution={"type": "uniform"},
            ),
        ],
    )
    mi_wrapper = check_scene_element(mixed.phase, mi.PhaseFunction)
    weight_1 = np.squeeze(mi_wrapper.parameters["weight.data"])
    weight_2 = np.squeeze(mi_wrapper.parameters["phase_1.weight.data"])

    middle = np.argwhere(geometry.zgrid.layers <= 50.0 * ureg.km).max() + 1
    fourfive = np.argwhere(geometry.zgrid.layers <= 80.0 * ureg.km).max() + 1

    assert np.all(weight_1[:middle] == 0.0)
    assert np.all(weight_1[middle:] == 1.0)
    assert np.all(weight_2[:fourfive] == 0.0)
    assert np.all(weight_2[fourfive:] == 1.0)

    # Third check: overlapping components
    # Component 1 has twice the optical thickness and extent of component 2,
    # therefore they have the same extinction coefficient
    mixed = HeterogeneousAtmosphere(
        geometry=geometry,
        particle_layers=[
            ParticleLayer(
                bottom=0.0 * ureg.km,
                top=100.0 * ureg.km,
                tau_ref=1.0,
                distribution={"type": "uniform"},
            ),
            ParticleLayer(
                bottom=50.0 * ureg.km,
                top=100.0 * ureg.km,
                tau_ref=0.5,
                distribution={"type": "uniform"},
            ),
        ],
    )
    mi_wrapper = check_scene_element(mixed.phase, mi.PhaseFunction)
    weights = np.squeeze(mi_wrapper.parameters["weight.data"])
    middle = np.argwhere(geometry.zgrid.layers <= 50.0 * ureg.km).max() + 1

    assert np.all(weights[:middle] == 0.0)
    assert np.all(weights[middle:] == 0.5)


@pytest.mark.slow
def test_heterogeneous_scale(mode_mono, atmosphere_us_standard_mono):
    atmosphere = HeterogeneousAtmosphere(
        geometry="plane_parallel",
        molecular_atmosphere=atmosphere_us_standard_mono,
        particle_layers=[ParticleLayer() for _ in range(2)],
        scale=2.0,
    )
    template, _ = traverse(atmosphere)
    assert template["medium_atmosphere.scale"] == 2.0

    # particles phae function is tabulated for wavelengths starting at 350 nm
    for si in atmosphere.spectral_set().spectral_indices():
        if si.w.m > 350:
            break

    kernel_context = KernelContext(si=si)

    # The scene element produces valid kernel dictionary specifications
    check_scene_element(atmosphere, ctx=kernel_context)


def test_heterogeneous_blend_switches(
    mode_mono,
    atmosphere_us_standard_mono,
):
    # Rayleigh-only atmosphere + particle layer combination works
    assert HeterogeneousAtmosphere(
        molecular_atmosphere=atmosphere_us_standard_mono,
        particle_layers=[ParticleLayer()],
    )


@pytest.mark.parametrize(
    "particle_radprops",
    ["particle_dataset_absorbing_only", "particle_dataset_scattering_only"],
)
def test_heterogeneous_absorbing_mol_atm(
    mode_ckd, particle_radprops, request, atmosphere_us_standard_ckd
):
    """
    Phase function weights are correct when the molecular atmosphere is
    absorbing-only and the particle layer is either absorbing-only or
    scattering-only.
    """
    # Expand fixture
    _particle_radprops = request.getfixturevalue(particle_radprops)
    # Create the heterogeneous atmosphere
    pl_bottom = 1.0 * ureg.km  # arbitrary
    pl_top = 4.0 * ureg.km  # arbitrary
    particle_layer = ParticleLayer(
        bottom=pl_bottom, top=pl_top, dataset=_particle_radprops
    )
    atmosphere = HeterogeneousAtmosphere(
        molecular_atmosphere=atmosphere_us_standard_ckd,
        particle_layers=particle_layer,
        geometry={
            "type": "spherical_shell",  # arbitrary
            "zgrid": np.linspace(0, 120, 121) * ureg.km,
        },
    )

    # Collect phase function weights
    mi_wrapper = check_scene_element(atmosphere.phase, mi.PhaseFunction)
    weights = np.squeeze(mi_wrapper.parameters["weight.volume.data"])

    # Extract phase function weights
    inside_particle_layer = (atmosphere.geometry.zgrid.layers >= pl_bottom) & (
        atmosphere.geometry.zgrid.layers <= pl_top
    )

    # Outside the particle layer, the phase function weight should be zero.
    assert np.all(weights[~inside_particle_layer] == 0.0)

    # Within the particle layer, the phase function weight should be:
    #   - zero, if the particle layer is not scattering (i.e., absorbing-only)
    #   - larger than zero, if the particle layer is scattering
    if particle_radprops == "particle_dataset_absorbing_only":
        assert np.all(weights[inside_particle_layer] == 0.0)
    elif particle_radprops == "particle_dataset_scattering_only":
        assert np.all(weights[inside_particle_layer] > 0.0)
    else:
        raise ValueError(
            f"Test parametrisation inconsistent. Expected 'absorbing_only' or "
            f"'scattering_only' (got {particle_radprops})"
        )



from pathlib import Path

import numpy as np
import pytest

import eradiate
from eradiate.experiments import CanopyExperiment
from eradiate.scenes.biosphere import load_scenario
from eradiate.scenes.biosphere._canopy_loader import apply_transformation
from eradiate.scenes.biosphere._rami_scenarios import (
    RAMIActualCanopies,
    RAMIHeterogeneousAbstractCanopies,
    RAMIHomogeneousAbstractCanopies,
    RAMIScenarioVersion,
    generate_name,
)
from eradiate.units import unit_registry as ureg

eradiate.set_mode("mono")


SAMPLE_SCENARIO_RPV = """{
  "surface": {
    "type": "rpv",
    "rho_0": {
        "wavelengths": [
          442.948,
          490.448,
          560.4304999999999,
          665.2445,
          681.556,
          709.1095,
          754.184,
          833.5,
          865.587,
          1242.5,
          1610.5,
          2120.0,
          2199.0
        ],
        "wavelengths_units": "nanometer",
        "values": [
          0.037938,
          0.04221,
          0.059786,
          0.088836,
          0.091155,
          0.096037,
          0.114102,
          0.120937,
          0.122402,
          0.147302,
          0.152672,
          0.152184,
          0.146692
        ],
        "type": "interpolated"
    },
    "g": {
        "wavelengths": [
          442.948,
          490.448,
          560.4304999999999,
          665.2445,
          681.556,
          709.1095,
          754.184,
          833.5,
          865.587,
          1242.5,
          1610.5,
          2120.0,
          2199.0
        ],
        "wavelengths_units": "nanometer",
        "values": [
          -0.25,
          -0.25,
          -0.25,
          -0.25,
          -0.25,
          -0.25,
          -0.2,
          -0.2,
          -0.2,
          -0.2,
          -0.2,
          -0.2,
          -0.2
        ],
        "type": "interpolated"
    },
    "k": {
        "wavelengths": [
          442.948,
          490.448,
          560.4304999999999,
          665.2445,
          681.556,
          709.1095,
          754.184,
          833.5,
          865.587,
          1242.5,
          1610.5,
          2120.0,
          2199.0
        ],
        "wavelengths_units": "nanometer",
        "values": [
          0.55,
          0.55,
          0.55,
          0.55,
          0.55,
          0.55,
          0.6,
          0.6,
          0.6,
          0.6,
          0.6,
          0.6,
          0.6
        ],
        "type": "interpolated"
    }
  },
  "canopy": {
    "type": "discrete_canopy",
    "instanced_canopy_elements": [
      {
        "canopy_element": {
          "type": "mesh_tree",
          "id": "Planophile",
          "mesh_tree_elements": [
            {
              "id": "Planophile",
              "mesh_filename": "tree-species/Planophile/Planophile.ply",
              "reflectance": {
                "quantity": "reflectance",
                "wavelengths": [
                  442.948,
                  490.448,
                  560.4304999999999,
                  665.2445,
                  681.556,
                  709.1095,
                  754.184,
                  833.5,
                  865.587,
                  1242.5,
                  1610.5,
                  2120.0,
                  2199.0
                ],
                "wavelengths_units": "nanometer",
                "values": [
                  0.053892,
                  0.057882,
                  0.136485,
                  0.055265,
                  0.052734,
                  0.214271,
                  0.4771,
                  0.494542,
                  0.496112,
                  0.461875,
                  0.332809,
                  0.158912,
                  0.181612
                ],
                "type": "interpolated"
              },
              "transmittance": {
                "quantity": "transmittance",
                "wavelengths": [
                  442.948,
                  490.448,
                  560.4304999999999,
                  665.2445,
                  681.556,
                  709.1095,
                  754.184,
                  833.5,
                  865.587,
                  1242.5,
                  1610.5,
                  2120.0,
                  2199.0
                ],
                "wavelengths_units": "nanometer",
                "values": [
                  0.005325,
                  0.010844,
                  0.088037,
                  0.015605,
                  0.01291,
                  0.169364,
                  0.422238,
                  0.440022,
                  0.44141,
                  0.43188,
                  0.340786,
                  0.185841,
                  0.22086
                ],
                "type": "interpolated"
              }
            }
          ]
        },
        "instance_positions": [
          [
            [
              1.0,
              0.0,
              0.0,
              12.5
            ],
            [
              0.0,
              1.0,
              0.0,
              12.5
            ],
            [
              0.0,
              0.0,
              1.0,
              0.0
            ],
            [
              0.0,
              0.0,
              0.0,
              1.0
            ]
          ]
        ]
      }
    ],
    "size": [
      25.0,
      25.0,
      1.0
    ]
  }
}
"""

SAMPLE_SCENARIO_SURFACE = """{
  "surface": {
    "type": "lambertian",
    "reflectance": {
      "quantity": "reflectance",
      "wavelengths": [
        442.95000000000005,
        490.45000000000005,
        560.4,
        665.25,
        681.55,
        709.1,
        754.15,
        833.5,
        865.6,
        1242.5,
        1610.5,
        2120.0,
        2199.0
      ],
      "wavelengths_units": "nanometer",
      "values": [
        0.882742,
        0.897852,
        0.907004,
        0.910396,
        0.913883,
        0.902941,
        0.900788,
        0.856648,
        0.848049,
        0.368409,
        0.04536,
        0.016815,
        0.053261
      ],
      "type": "interpolated"
    }
  },
  "canopy": {
    "type": "discrete_canopy",
    "instanced_canopy_elements": [
      {
        "canopy_element": {
          "type": "mesh_tree",
          "id": "PIMO1",
          "mesh_tree_elements": [
            {
              "id": "Wood",
              "mesh_filename": "tree-species/Pinus-Montana/PIMO1/Pinus_mugo_PIMO1_-_Original/Wood.ply",
              "reflectance": {
                "quantity": "reflectance",
                "wavelengths": [
                  442.95000000000005,
                  490.45000000000005,
                  560.4,
                  665.25,
                  681.55,
                  709.1,
                  754.15,
                  833.5,
                  865.6,
                  1242.5,
                  1610.5,
                  2120.0,
                  2199.0
                ],
                "wavelengths_units": "nanometer",
                "values": [
                  0.072764,
                  0.082621,
                  0.109325,
                  0.157819,
                  0.167218,
                  0.200941,
                  0.239725,
                  0.303587,
                  0.329257,
                  0.525503,
                  0.511729,
                  0.383173,
                  0.394486
                ],
                "type": "interpolated"
              }
            }
          ]
        },
        "instance_positions": [
          [
            [
              0.2813481616359169,
              0.9596057586030265,
              0.0,
              18.5271
            ],
            [
              -0.9596057586030265,
              0.2813481616359169,
              0.0,
              97.4987
            ],
            [
              0.0,
              0.0,
              1.0,
              0.0
            ],
            [
              0.0,
              0.0,
              0.0,
              1.0
            ]
          ]
        ]
      }
    ],
    "size": [
      103.1214,
      103.2308,
      15.0213
    ]
  }
}
"""


def generate_expected_output_rpv(mesh_filename: Path):
    return {
        "canopy": {
            "instanced_canopy_elements": [
                {
                    "canopy_element": {
                        "id": "Planophile",
                        "mesh_tree_elements": [
                            {
                                "id": "Planophile",
                                "mesh_filename": mesh_filename,
                                "reflectance": {
                                    "quantity": "reflectance",
                                    "type": "interpolated",
                                    "values": pytest.approx(
                                        [
                                            0.053892,
                                            0.057882,
                                            0.136485,
                                            0.055265,
                                            0.052734,
                                            0.214271,
                                            0.4771,
                                            0.494542,
                                            0.496112,
                                            0.461875,
                                            0.332809,
                                            0.158912,
                                            0.181612,
                                        ]
                                    ),
                                    "wavelengths": pytest.approx(
                                        [
                                            442.948,
                                            490.448,
                                            560.43045,
                                            665.2445,
                                            681.556,
                                            709.1095,
                                            754.184,
                                            833.5,
                                            865.587,
                                            1242.5,
                                            1610.5,
                                            2120.0,
                                            2199.0,
                                        ]
                                    ),
                                    "wavelengths_units": "nanometer",
                                },
                                "transmittance": {
                                    "quantity": "transmittance",
                                    "type": "interpolated",
                                    "values": pytest.approx(
                                        [
                                            0.005325,
                                            0.010844,
                                            0.088037,
                                            0.015605,
                                            0.01291,
                                            0.169364,
                                            0.422238,
                                            0.440022,
                                            0.44141,
                                            0.43188,
                                            0.340786,
                                            0.185841,
                                            0.22086,
                                        ]
                                    ),
                                    "wavelengths": pytest.approx(
                                        [
                                            442.948,
                                            490.448,
                                            560.43045,
                                            665.2445,
                                            681.556,
                                            709.1095,
                                            754.184,
                                            833.5,
                                            865.587,
                                            1242.5,
                                            1610.5,
                                            2120.0,
                                            2199.0,
                                        ]
                                    ),
                                    "wavelengths_units": "nanometer",
                                },
                            }
                        ],
                        "type": "mesh_tree",
                    },
                    "instance_positions": [pytest.approx([0.0, 0.0, 0.0])],
                }
            ],
            "padding": 0,
            "size": pytest.approx([25.0, 25.0, 1.0]),
            "type": "discrete_canopy",
        },
        "surface": {
            "g": {
                "type": "interpolated",
                "values": pytest.approx(
                    [
                        -0.25,
                        -0.25,
                        -0.25,
                        -0.25,
                        -0.25,
                        -0.25,
                        -0.2,
                        -0.2,
                        -0.2,
                        -0.2,
                        -0.2,
                        -0.2,
                        -0.2,
                    ]
                ),
                "wavelengths": pytest.approx(
                    [
                        442.948,
                        490.448,
                        560.4305,
                        665.2445,
                        681.556,
                        709.1095,
                        754.184,
                        833.5,
                        865.587,
                        1242.5,
                        1610.5,
                        2120.0,
                        2199.0,
                    ]
                ),
                "wavelengths_units": "nanometer",
            },
            "k": {
                "type": "interpolated",
                "values": pytest.approx(
                    [
                        0.55,
                        0.55,
                        0.55,
                        0.55,
                        0.55,
                        0.55,
                        0.6,
                        0.6,
                        0.6,
                        0.6,
                        0.6,
                        0.6,
                        0.6,
                    ]
                ),
                "wavelengths": pytest.approx(
                    [
                        442.948,
                        490.448,
                        560.4305,
                        665.2445,
                        681.556,
                        709.1095,
                        754.184,
                        833.5,
                        865.587,
                        1242.5,
                        1610.5,
                        2120.0,
                        2199.0,
                    ]
                ),
                "wavelengths_units": "nanometer",
            },
            "rho_0": {
                "type": "interpolated",
                "values": pytest.approx(
                    [
                        0.037938,
                        0.04221,
                        0.059786,
                        0.088836,
                        0.091155,
                        0.096037,
                        0.114102,
                        0.120937,
                        0.122402,
                        0.147302,
                        0.152672,
                        0.152184,
                        0.146692,
                    ]
                ),
                "wavelengths": pytest.approx(
                    [
                        442.948,
                        490.448,
                        560.4305,
                        665.2445,
                        681.556,
                        709.1095,
                        754.184,
                        833.5,
                        865.587,
                        1242.5,
                        1610.5,
                        2120.0,
                        2199.0,
                    ]
                ),
                "wavelengths_units": "nanometer",
            },
            "type": "rpv",
        },
    }


def generate_expected_output_surface(mesh_filename: Path):
    return {
        "canopy": {
            "type": "discrete_canopy",
            "instanced_canopy_elements": [
                {
                    "canopy_element": {
                        "type": "mesh_tree",
                        "id": "PIMO1",
                        "mesh_tree_elements": [
                            {
                                "id": "Wood",
                                "mesh_filename": mesh_filename,
                                "reflectance": {
                                    "quantity": "reflectance",
                                    "wavelengths": pytest.approx(
                                        [
                                            442.95,
                                            490.45,
                                            560.4,
                                            665.25,
                                            681.55,
                                            709.1,
                                            754.15,
                                            833.5,
                                            865.6,
                                            1242.5,
                                            1610.5,
                                            2120.0,
                                            2199.0,
                                        ]
                                    ),
                                    "wavelengths_units": "nanometer",
                                    "values": pytest.approx(
                                        [
                                            0.072764,
                                            0.082621,
                                            0.109325,
                                            0.157819,
                                            0.167218,
                                            0.200941,
                                            0.239725,
                                            0.303587,
                                            0.329257,
                                            0.525503,
                                            0.511729,
                                            0.383173,
                                            0.394486,
                                        ],
                                    ),
                                    "type": "interpolated",
                                },
                            }
                        ],
                    },
                    "instance_positions": [pytest.approx([-33.0336, 45.8833, 0.0])],
                }
            ],
            "padding": 0,
            "size": pytest.approx([103.1214, 103.2308, 15.0213]),
        },
        "surface": {
            "type": "lambertian",
            "reflectance": {
                "quantity": "reflectance",
                "wavelengths": pytest.approx(
                    [
                        442.95,
                        490.45,
                        560.4,
                        665.25,
                        681.55,
                        709.1,
                        754.15,
                        833.5,
                        865.6,
                        1242.5,
                        1610.5,
                        2120.0,
                        2199.0,
                    ]
                ),
                "wavelengths_units": "nanometer",
                "values": pytest.approx(
                    [
                        0.882742,
                        0.897852,
                        0.907004,
                        0.910396,
                        0.913883,
                        0.902941,
                        0.900788,
                        0.856648,
                        0.848049,
                        0.368409,
                        0.04536,
                        0.016815,
                        0.053261,
                    ]
                ),
                "type": "interpolated",
            },
        },
    }


class TestApplyTransformation:
    """Test case for the apply_transformation function."""

    @staticmethod
    def test_transformation():
        """Test the apply_transformation function with a specific transformation matrix and center point."""
        transf = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
        center = np.array([1.0, 2.0, 0.0]) / 2
        expected_output = np.array([-0.5, -1.0, 0.0])
        assert np.allclose(apply_transformation(transf, center), expected_output)


class TestLoadScenario:
    """Test cases for the load_scenario function."""

    @staticmethod
    def test_load_scenario_rpv_without_spectral_data(tmp_path: Path):
        """Tests loading an RPV scenario without additional spectral data applied to the scenario elements."""
        scenario_folder = tmp_path

        (scenario_folder / "scenario.json").write_text(SAMPLE_SCENARIO_RPV)
        planophile_file = (
            scenario_folder / "tree-species" / "Planophile" / "Planophile.ply"
        )
        planophile_file.parent.mkdir(parents=True)
        planophile_file.touch()

        expected_output = generate_expected_output_rpv(planophile_file)
        parsed_scene = load_scenario(scenario_folder, padding=0)

        assert parsed_scene == expected_output

        # Check that the scene can be used to create an experiment
        CanopyExperiment(
            **parsed_scene,
            measures={
                "type": "perspective",
                "film_resolution": (10, 10),
                "origin": [1.0, 0.0, 0.0],
                "target": [0.0, 0.0, 0.0],
                "up": [0.0, 0.0, 1.0],
                "sampler": "ldsampler",
                "fov": 20.0,
                "spp": 4,
                "srf": {
                    "type": "multi_delta",
                    "wavelengths": np.array([660, 550, 440]) * ureg.nm,
                },
            },
            illumination={"type": "directional", "zenith": 0.0, "azimuth": 0.0},
        )

    @staticmethod
    def test_load_scenario_surface_without_spectral_data(tmp_path: Path):
        """Tests loading a surface scenario without additional spectral data applied to the scenario elements."""
        scenario_folder = tmp_path

        (scenario_folder / "scenario.json").write_text(SAMPLE_SCENARIO_SURFACE)
        tree_file = (
            scenario_folder
            / "tree-species"
            / "Pinus-Montana"
            / "PIMO1"
            / "Pinus_mugo_PIMO1_-_Original"
            / "Wood.ply"
        )
        tree_file.parent.mkdir(parents=True)
        tree_file.touch()

        expected_output = generate_expected_output_surface(tree_file)

        parsed_scene = load_scenario(scenario_folder, padding=0)

        assert parsed_scene == expected_output
        # Check that the scene can be used to create an experiment
        CanopyExperiment(
            **parsed_scene,
            measures={
                "type": "perspective",
                "film_resolution": (10, 10),
                "origin": [1.0, 0.0, 0.0],
                "target": [0.0, 0.0, 0.0],
                "up": [0.0, 0.0, 1.0],
                "sampler": "ldsampler",
                "fov": 20.0,
                "spp": 4,
                "srf": {
                    "type": "multi_delta",
                    "wavelengths": np.array([660, 550, 440]) * ureg.nm,
                },
            },
            illumination={"type": "directional", "zenith": 0.0, "azimuth": 0.0},
        )

    @staticmethod
    def test_load_scenario_rpv_with_spectral_data(tmp_path: Path):
        """Tests loading an RPV scenario with additional spectral data applied to the scenario elements."""
        scenario_folder = tmp_path

        (scenario_folder / "scenario.json").write_text(SAMPLE_SCENARIO_RPV)
        elem_file = scenario_folder / "tree-species" / "Planophile" / "Planophile.ply"
        elem_file.parent.mkdir(parents=True)
        elem_file.touch()

        expected_output = generate_expected_output_rpv(elem_file)

        spectral_data = {
            "Planophile": {
                "Planophile": {
                    "reflectance": (
                        {
                            "quantity": "reflectance",
                            "wavelengths": [
                                442.948,
                                490.448,
                                560.43045,
                                665.2445,
                                681.556,
                                709.1095,
                                754.184,
                                833.5,
                                865.587,
                                1242.5,
                                1610.5,
                                2120.0,
                                2199.0,
                            ],
                            "wavelengths_units": "nanometer",
                            "values": [
                                0.053892,
                                0.057882,
                                0.136485,
                                0.055265,
                                0.052734,
                                0.214271,
                                0.4771,
                                0.494542,
                                0.496112,
                                0.461875,
                                0.332809,
                                0.158912,
                                0.181612,
                            ],
                            "type": "interpolated",
                        }
                    ),
                },
            },
        }
        del expected_output["canopy"]["instanced_canopy_elements"][0]["canopy_element"][
            "mesh_tree_elements"
        ][0]["transmittance"]
        parsed_scene = load_scenario(
            scenario_folder, padding=0, spectral_data=spectral_data
        )

        assert parsed_scene == expected_output

        # Check that the scene can be used to create an experiment
        CanopyExperiment(
            **parsed_scene,
            measures={
                "type": "perspective",
                "film_resolution": (10, 10),
                "origin": [1.0, 0.0, 0.0],
                "target": [0.0, 0.0, 0.0],
                "up": [0.0, 0.0, 1.0],
                "sampler": "ldsampler",
                "fov": 20.0,
                "spp": 4,
                "srf": {
                    "type": "multi_delta",
                    "wavelengths": np.array([660, 550, 440]) * ureg.nm,
                },
            },
            illumination={"type": "directional", "zenith": 0.0, "azimuth": 0.0},
        )

    @staticmethod
    def test_load_scenario_missing_scenario(tmp_path: Path):
        """Tests loading a scenario that does not exist."""
        scenario_folder = tmp_path
        padding = 0
        with pytest.raises(FileNotFoundError):
            load_scenario(scenario_folder, padding)


class TestGenerateName:
    @staticmethod
    def test_generate_name_actual_canopies_original():
        assert generate_name(RAMIActualCanopies.JARVSELJA_PINE_STAND) == "HET07_JPS_SUM"

    @staticmethod
    def test_generate_name_actual_canopies_simplified():
        assert (
            generate_name(
                RAMIActualCanopies.JARVSELJA_PINE_STAND, RAMIScenarioVersion.SIMPLIFIED
            )
            == "HET07_JPS_SUM-simplified"
        )

    @staticmethod
    def test_generate_name_heterogeneous_abstract_canopies_original():
        assert (
            generate_name(
                RAMIHeterogeneousAbstractCanopies.ANISOTROPIC_BACKGROUND_OVERSTOREY_SPARSE_BRF_MODEL_A
            )
            == "HET10_DIS_S1A"
        )

    @staticmethod
    def test_generate_name_heterogeneous_abstract_canopies_simplified():
        assert (
            generate_name(
                RAMIHeterogeneousAbstractCanopies.ANISOTROPIC_BACKGROUND_OVERSTOREY_SPARSE_BRF_MODEL_A,
                RAMIScenarioVersion.SIMPLIFIED,
            )
            == "HET10_DIS_S1A-simplified"
        )

    @staticmethod
    def test_generate_name_homogeneous_abstract_canopies_original():
        assert (
            generate_name(
                RAMIHomogeneousAbstractCanopies.ANISOTROPIC_BACKGROUND_PLANOPHILE_A
            )
            == "HOM23_DIS_P1A"
        )

    @staticmethod
    def test_generate_name_homogeneous_abstract_canopies_simplified():
        assert (
            generate_name(
                RAMIHomogeneousAbstractCanopies.ANISOTROPIC_BACKGROUND_PLANOPHILE_A,
                RAMIScenarioVersion.SIMPLIFIED,
            )
            == "HOM23_DIS_P1A-simplified"
        )



import mitsuba as mi

from eradiate import unit_registry as ureg
from eradiate.scenes.illumination import ConstantIllumination
from eradiate.scenes.spectra import UniformSpectrum
from eradiate.test_tools.types import check_scene_element


def test_constant_construct(modes_all):
    # Construction without argument succeeds
    illumination = ConstantIllumination()
    assert illumination
    assert isinstance(illumination.radiance, UniformSpectrum)
    assert illumination.radiance.value == 1.0 * ureg("W/m^2/sr/nm")


def test_constant_kernel_dict(modes_all_double):
    # The associated kernel dict is correctly formed and can be loaded
    illumination = ConstantIllumination()
    check_scene_element(illumination, mi.Emitter)



import mitsuba as mi
import numpy as np

import eradiate.data
from eradiate import unit_registry as ureg
from eradiate.scenes.illumination import SpotIllumination
from eradiate.test_tools.types import check_scene_element


def test_construct_basic(mode_mono):
    # Construct without parameters
    illumination = SpotIllumination()
    check_scene_element(illumination, mi_cls=mi.Emitter)

    # Define all parameters
    illumination = SpotIllumination(
        target=[0, 0, 0],
        origin=[1, 1, 1],
        up=[1, 0, 1],
        beam_width=10 * ureg.deg,
        intensity=10,
    )
    check_scene_element(illumination, mi_cls=mi.Emitter)


def test_construct_texture(mode_mono, tmp_path):
    # First we set up a temporary directory with a bitmap file
    from PIL import Image

    filename = tmp_path / "texture.bmp"
    array = np.ones((3, 3, 3))
    im = Image.fromarray(np.uint8(array * 255))
    im.save(filename)

    # Construct with custom beam profile filename
    illumination = SpotIllumination(
        target=[0, 0, 0],
        origin=[1, 1, 1],
        up=[1, 0, 1],
        beam_width=10 * ureg.deg,
        intensity=10,
        beam_profile=str(filename),
    )
    check_scene_element(illumination, mi_cls=mi.Emitter)

    # Construct with beam profile from data store
    illumination = SpotIllumination(
        target=[0, 0, 0],
        origin=[1, 1, 1],
        up=[1, 0, 1],
        beam_width=10 * ureg.deg,
        intensity=10,
        beam_profile=eradiate.data.data_store.fetch("textures/gaussian_3sigma.bmp"),
    )
    check_scene_element(illumination, mi_cls=mi.Emitter)


def test_construct_from_size(mode_mono):
    spot_radius = 1 * ureg.m
    beam_angle = 3 * ureg.deg

    # spot_radius / tan(beam_angle/2.)
    expected_distance = 38.18846

    illumination = SpotIllumination.from_size_at_target(
        target=[0, 0, 0],
        direction=[0, 0, -1],
        spot_radius=spot_radius,
        beam_width=beam_angle,
    )

    assert np.allclose(illumination.origin.m_as(ureg.m), [0, 0, expected_distance])

    check_scene_element(illumination, mi_cls=mi.Emitter)



import mitsuba as mi
import pytest

from eradiate.scenes.bsdfs import RTLSBSDF
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"f_iso": 0.3, "f_vol": 1.4, "f_geo": -0.23},
        {
            "f_iso": 0.3,
            "f_vol": {"type": "uniform", "value": 0.3},
            "f_geo": {
                "type": "interpolated",
                "wavelengths": [300.0, 800.0],
                "values": [-0.23, 0.23],
            },
            "h": 0.2,
        },
    ],
    ids=["noargs", "uniform", "mixed"],
)
def test_rpv_construct(modes_all, kwargs):
    # Default constructor
    assert RTLSBSDF(**kwargs)


def test_rtls(modes_all_double):
    bsdf = RTLSBSDF(
        f_iso=0.3,
        f_geo={"type": "uniform", "value": 0.3},
        f_vol={
            "type": "interpolated",
            "wavelengths": [300.0, 800.0],
            "values": [-0.23, 0.23],
        },
        h=0.2,
        r=0.1,
        b=0.97,
    )

    check_scene_element(bsdf, mi.BSDF)



import attr
import pytest

from eradiate import unit_registry as ureg
from eradiate.validators import on_quantity


def test_on_quantity():
    v = on_quantity(attr.validators.instance_of(float))

    # This should succeed
    v(None, None, 1.0)
    v(None, None, ureg.Quantity(1.0, "km"))

    # This should fail
    @attr.s
    class Attribute:  # Tiny class to pass an appropriate attribute argument
        name = attr.ib()

    attribute = Attribute(name="attribute")

    with pytest.raises(TypeError):
        v(None, attribute, "1.")
    with pytest.raises(TypeError):
        v(None, attribute, ureg.Quantity("1.", "km"))



__docformat__ = "reStructuredText"

from docutils import nodes
from docutils.parsers.rst.directives.tables import Table
from docutils.statemachine import ViewList
from docutils.utils import SystemMessagePropagation
from sphinx.util.nodes import nested_parse_with_titles


class PluginParameters(Table):
    """
    Implement tables whose data is encoded as a uniform two-level bullet list
    using to describe Mitsuba 3 plugin parameters.
    """

    optional_arguments = 1
    option_spec = {"extra-rows": int}

    def run(self):

        extra_rows = self.options.get("extra-rows", 0)
        if not self.content:
            error = self.state_machine.reporter.error(
                'The "%s" directive is empty; content required.' % self.name,
                nodes.literal_block(self.block_text, self.block_text),
                line=self.lineno,
            )
            return [error]
        title, messages = self.make_title()
        node = nodes.Element()  # anonymous container for parsing
        self.state.nested_parse(self.content, self.content_offset, node)

        num_cols = self.check_list_content(node)

        # Hardcode this:
        col_widths = [20, 12, 60, 8]

        table_data = [[item.children for item in row_list[0]] for row_list in node[0]]
        header_rows = self.options.get("header-rows", 1)
        stub_columns = self.options.get("stub-columns", 0)
        self.check_table_dimensions(table_data, header_rows - 1, stub_columns)

        table_node = self.build_table_from_list(
            table_data, col_widths, header_rows, extra_rows, stub_columns
        )
        table_node["classes"] += self.options.get("class", ["paramstable"])
        self.add_name(table_node)
        if title:
            table_node.insert(0, title)
        return [table_node] + messages

    def check_list_content(self, node):
        if len(node) != 1 or not isinstance(node[0], nodes.bullet_list):
            error = self.state_machine.reporter.error(
                'Error parsing content block for the "%s" directive: '
                "exactly one bullet list expected." % self.name,
                nodes.literal_block(self.block_text, self.block_text),
                line=self.lineno,
            )
            raise SystemMessagePropagation(error)
        list_node = node[0]
        # Check for a uniform two-level bullet list:
        for item_index in range(len(list_node)):
            item = list_node[item_index]
            if len(item) != 1 or not isinstance(item[0], nodes.bullet_list):
                error = self.state_machine.reporter.error(
                    'Error parsing content block for the "%s" directive: '
                    "two-level bullet list expected, but row %s does not "
                    "contain a second-level bullet list." % (self.name, item_index + 1),
                    nodes.literal_block(self.block_text, self.block_text),
                    line=self.lineno,
                )
                raise SystemMessagePropagation(error)
            else:
                num_cols = len(item[0])
        return 4

    def build_table_from_list(
        self, table_data, col_widths, header_rows, extra_rows, stub_columns
    ):
        table = nodes.table()
        tgroup = nodes.tgroup(cols=len(col_widths))
        table += tgroup
        for col_width in col_widths:
            colspec = nodes.colspec(colwidth=col_width)
            if stub_columns:
                colspec.attributes["stub"] = 1
                stub_columns -= 1
            tgroup += colspec
        rows = []

        # Append first row
        header_text = ["Parameter", "Type", "Description", "Flags"]
        header_row_node = nodes.row()
        for text in header_text:
            entry = nodes.entry()
            entry += [nodes.paragraph(text=text)]
            header_row_node += entry
        rows.append(header_row_node)

        for row in table_data:
            row_node = nodes.row()
            for i, cell in enumerate(row):
                entry = nodes.entry()

                # force the first column to be write in paramtype style
                if i == 0:
                    rst = ViewList()
                    rst.append(
                        """:paramtype:`{name}`""".format(name=str(cell[0][0])), "", 0
                    )
                    parsed_node = nodes.section()
                    parsed_node.document = self.state.document
                    nested_parse_with_titles(self.state, rst, parsed_node)

                    entry += [parsed_node[0]]
                else:
                    entry += cell

                row_node += entry

            # Add an empty cell if only 3 entries are provided
            if len(row) == 3:
                entry = nodes.entry()
                entry += [nodes.paragraph(text="")]
                row_node += entry

            rows.append(row_node)
        if header_rows:
            thead = nodes.thead()
            thead.extend(rows[:header_rows])
            tgroup += thead
        tbody = nodes.tbody()

        if extra_rows > 0:
            tbody.extend(rows[header_rows : extra_rows + 1])
        else:
            tbody.extend(rows[header_rows:])
        tgroup += tbody

        # Handle the extra state header and rows
        if extra_rows > 0:
            # Append first row
            header_text = ["State parameters", "", "", ""]
            header_row_node = nodes.row()
            for text in header_text:
                entry = nodes.entry()
                entry += [nodes.paragraph(text=text)]
                header_row_node += entry

            thead = nodes.thead()
            thead.extend([header_row_node])
            tgroup += thead

            tbody = nodes.tbody()
            tbody.extend(rows[extra_rows + 1 :])
            tgroup += tbody

        return table


def setup(app):
    app.add_directive("pluginparameters", PluginParameters)
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import attr

from eradiate._factory import Factory
from eradiate.units import unit_registry as ureg


def test_factory_convert():
    factory = Factory()

    # We register a type with a nondefault constructor used for dict-based
    # creation
    @factory.register(type_id="mycls", dict_constructor="foo")
    @attr.s
    class MyClass:
        field = attr.ib(default=None)

        @classmethod
        def foo(cls):
            return cls(field="foo")

        @classmethod
        def bar(cls):
            return cls(field="bar")

    # MyClass instances are not modified
    o = MyClass()
    assert factory.convert(o) is o

    # Dict conversion uses the specified constructor
    assert factory.convert({"type": "mycls"}) == MyClass(field="foo")

    # We can override the dict constructor, either to use the default one...
    assert factory.convert({"type": "mycls", "construct": None}) == MyClass()
    # ... or a custom one
    assert factory.convert({"type": "mycls", "construct": "bar"}) == MyClass(
        field="bar"
    )

    # Units are attached to fields prior to conversion
    assert factory.convert(
        {"type": "mycls", "construct": None, "field": 1.0, "field_units": "m"}
    ) == MyClass(field=1.0 * ureg.m)



import os

import attrs
import matplotlib.pyplot as plt
import numpy as np
import pytest
import xarray as xr

import eradiate
from eradiate import unit_registry as ureg


def compute(ray_offset="distant", sigma=1.0, rho=1.0, exp_cls="AtmosphereExperiment"):
    result = []
    spps = [4**i for i in range(0, 11)]

    for spp in spps:
        exp = getattr(eradiate.experiments, exp_cls)(
            geometry={"type": "plane_parallel", "toa_altitude": 1.0 * ureg.m},
            atmosphere={
                "type": "homogeneous",
                "sigma_a": sigma * ureg("m^-1"),
                "sigma_s": 0.0,
            },
            illumination={"type": "directional", "irradiance": 1.0},
            surface={"type": "lambertian", "reflectance": rho},
            measures={
                "type": "mdistant",
                "construct": "hplane",
                "zeniths": 0.0,
                "azimuth": 0.0,
                "ray_offset": None,
                "spp": spp,
            },
        )

        if ray_offset != "distant":
            exp = attrs.evolve(
                exp,
                measures={
                    "type": "mdistant",
                    "construct": "hplane",
                    "zeniths": 0.0,
                    "azimuth": 0.0,
                    "ray_offset": ray_offset * ureg.m,
                    "spp": spp,
                },
            )

        result.append(eradiate.run(exp).radiance.squeeze(drop=True))

    return xr.concat(result, dim="spp").assign_coords(spp=spps)


@pytest.mark.parametrize(
    "exp_cls", ["AtmosphereExperiment", "CanopyAtmosphereExperiment"]
)
def test_mdistant_insitu(artefact_dir, mode_mono, exp_cls):
    r"""
    In-situ multi-distant sensor
    ============================

    This test checks that the in-situ multi-distant sensor feature is correctly
    implemented. The test is performed for the following experiment classes:

    - ``AtmosphereExperiment``
    - ``CanopyAtmosphereExperiment``

    Rationale
    ---------

    - Surface: a Lambertian surface reflectance :math:`\rho = 1`.
    - Atmosphere: a homogeneous atmosphere with no scattering, an absorption
      coefficient :math:`\sigma = 1` and a thickness equal to 1.
    - Illumination: a directional light source at the zenith with radiance
      :math:`L_\mathrm{i} = 1.0`.
    - Sensor: an ``mdistant`` sensor targeting (0, 0, 0) in a single direction
      [0, 0, -1]. The ray offset is set to distant and a number of values
      between 0 and 1.

    Expected behaviour
    ------------------

    The recorded radiance is expected to be equal to
    :math:`\frac{1}{\pi} \exp \left( - \sigma (l + \mathrm{offset}) \right)`.
    """

    sigma = 1.0
    computed = {}
    expected = {}
    offsets = ["distant", 0.99, 0.5, 0.01]

    for offset in offsets:
        computed[offset] = compute(offset, sigma=sigma, exp_cls=exp_cls)
        expected[offset] = (
            np.exp(-sigma * (1.0 + (offset if offset != "distant" else 1.0))) / np.pi
        )

    # Plot results
    plt.figure()
    for offset in offsets:
        computed[offset].plot(ls=":", marker=".", xscale="log", label=f"{offset}")
        plt.axhline(expected[offset], zorder=0, c="whitesmoke", ls="--")

    plt.legend(ncol=2)
    test_name = f"test_mdistant_insitu-{exp_cls}"
    plt.title(test_name)

    outdir = os.path.join(artefact_dir, "plots")
    os.makedirs(outdir, exist_ok=True)
    filename = os.path.join(outdir, f"{test_name}.png")
    plt.savefig(filename, dpi=200, bbox_inches="tight")

    plt.close()

    # Actual test
    result = np.squeeze([computed[offset].isel(spp=-1) for offset in offsets])
    expected = np.squeeze([expected[offset] for offset in offsets])

    assert np.allclose(
        result, expected, rtol=1e-2
    ), f"{result = }, {expected = }\nPlot file: '{filename}'"



import mitsuba as mi
import pytest

from eradiate.scenes.integrators import (
    PathIntegrator,
    VolPathIntegrator,
    VolPathMISIntegrator,
)
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "integrator_cls",
    [PathIntegrator, VolPathIntegrator, VolPathMISIntegrator],
    ids=["path", "volpath", "volpathmis"],
)
@pytest.mark.parametrize(
    "kwargs",
    [{}, {"max_depth": 5, "rr_depth": 3, "hide_emitters": False}],
    ids=["noargs", "args"],
)
def test_path_tracers_construct(modes_all, integrator_cls, kwargs):
    # Construction succeed
    assert integrator_cls(**kwargs)


@pytest.mark.parametrize(
    "integrator_cls, kwargs",
    [
        (PathIntegrator, {"max_depth": 5, "rr_depth": 3, "hide_emitters": False}),
        (VolPathIntegrator, {"max_depth": 5, "rr_depth": 3, "hide_emitters": False}),
        (
            VolPathMISIntegrator,
            {
                "max_depth": 5,
                "rr_depth": 3,
                "hide_emitters": False,
                "use_spectral_mis": True,
            },
        ),
    ],
    ids=["path", "volpath", "volpathmis"],
)
def test_path_tracers_kernel_dict(mode_mono, integrator_cls, kwargs):
    integrator = integrator_cls(**kwargs)
    check_scene_element(integrator, mi.MonteCarloIntegrator)



import mitsuba as mi
import pytest

from eradiate import KernelContext
from eradiate.scenes.core import traverse
from eradiate.scenes.measure import MultiRadiancemeterMeasure
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "tested",
    [{}, dict(origins=[[0, 0, 0]] * 3, directions=[[1, 0, 0], [0, 1, 0], [0, 0, 1]])],
    ids=[
        "no_args",
        "origins_directions",
    ],
)
def test_multi_radiancemeter(modes_all_double, tested):
    measure = MultiRadiancemeterMeasure(**tested)
    check_scene_element(measure, mi.Sensor)


def test_multi_radiancemeter_medium(mode_mono):
    measure = MultiRadiancemeterMeasure()
    template, _ = traverse(measure)

    kdict = template.render(ctx=KernelContext())
    assert "medium" not in kdict

    kdict = template.render(
        ctx=KernelContext(kwargs={"measure.atmosphere_medium_id": "test_atmosphere"})
    )
    assert kdict["medium"] == {"type": "ref", "id": "test_atmosphere"}



import drjit as dr
import mitsuba as mi
import numpy as np
import pytest

from eradiate import KernelContext
from eradiate import unit_context_config as ucc
from eradiate import unit_registry as ureg
from eradiate.scenes.core import traverse
from eradiate.scenes.shapes import RectangleShape
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "kwargs, expected_reflectance",
    [({}, None), ({"bsdf": {"type": "lambertian", "reflectance": 0.0}}, 0.0)],
    ids=["noargs", "args"],
)
def test_rectangle_construct_kernel_dict(modes_all, kwargs, expected_reflectance):
    rectangle = RectangleShape(**kwargs)
    mi_wrapper = check_scene_element(rectangle, mi.Shape)

    if expected_reflectance is not None:
        assert mi_wrapper.parameters["bsdf.reflectance.value"] == expected_reflectance


def test_rectangle_construct_trafo(modes_all):
    assert RectangleShape(to_world=mi.Transform4f.scale(2))


@pytest.mark.parametrize(
    "kwargs, expected_transform",
    [
        (
            {"edges": [2, 4]},
            [(-1, -2, 0), (1, 2, 0)],
        ),
        (
            {"edges": [2, 2], "center": [0, 0, 1]},
            [(-1, -1, 1), (1, 1, 1)],
        ),
        (
            {"edges": [2, 2], "up": [-1, 0, 0]},
            [(1, -1, 0), (-1, 1, 0)],
        ),
        (
            {"edges": [2, 2], "normal": [0, -1, 0], "up": [-1, 0, 0]},
            [(1, 0, -1), (-1, 0, 1)],
        ),
        (
            {"to_world": True},
            [(0, 0, 1), (2, 2, 1)],
        ),
        (
            {
                "edges": [2, 4],
                "normal": [0, -1, 0],
                "up": [-1, 0, 0],
                "center": [0, 0, 1],
            },
            [(2, 0, 0), (-2, 0, 2)],
        ),
        (
            {
                "edges": [2, 4],
                "normal": [0, -1, 0],
                "up": [-1, 0, 0],
                "center": [0, 0, 1],
                "to_world": True,
            },
            [(0, 0, 1), (2, 2, 1)],
        ),
    ],
    ids=["edges", "center", "up", "normal_up", "to_world-only", "no-to_world", "full"],
)
def test_rectangle_params(mode_mono_double, kwargs, expected_transform):
    if "to_world" in kwargs:
        trafo = kwargs.pop("to_world")
        if trafo:
            to_world = mi.ScalarTransform4f.translate((1, 1, 1))
            rectangle = RectangleShape(**kwargs, to_world=to_world)
    else:
        rectangle = RectangleShape(**kwargs)

    # Set edges
    template, _ = traverse(rectangle)
    kernel_dict = template.render(ctx=KernelContext())
    to_world = kernel_dict["to_world"]
    assert dr.allclose(
        to_world.transform_affine(mi.Point3f(-1, -1, 0)), expected_transform[0]
    )
    assert dr.allclose(
        to_world.transform_affine(mi.Point3f(1, 1, 0)), expected_transform[1]
    )


def test_rectangle_surface():
    # Check that parameters are correctly specified
    with ucc.override(length="km"):
        surface = RectangleShape.surface(altitude=1.0, width=100.0)
        assert np.allclose(surface.center, [0, 0, 1] * ureg.km)
        assert np.allclose(surface.edges, [100, 100] * ureg.km)



import drjit as dr
import mitsuba as mi
import numpy as np
import pytest

from eradiate import unit_registry as ureg
from eradiate.scenes.core import traverse
from eradiate.scenes.measure import DistantFluxMeasure
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "tested",
    [
        {},
        {"target": [0, 0, 0]},
        {"target": {"type": "rectangle", "xmin": 0, "xmax": 1, "ymin": 0, "ymax": 1}},
    ],
    ids=[
        "no_args",
        "target_point",
        "target_rectangle",
    ],
)
def test_distant_flux_construct(modes_all_double, tested):
    measure = DistantFluxMeasure(**tested)
    check_scene_element(measure, mi_cls=mi.Sensor)


@pytest.mark.parametrize(
    ["direction", "frame"],
    [
        (
            [1, 0, 0],
            [[0, 0, -1], [0, 1, 0], [1, 0, 0]],
        ),
        (
            [0, 0, 1],
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        ),
    ],
)
def test_distant_flux_direction(modes_all_double, direction, frame):
    measure = DistantFluxMeasure(direction=direction)
    template, _ = traverse(measure)
    to_world = template["to_world"]

    # The reference frame is rotated as expected
    assert dr.allclose(to_world.transform_affine(mi.Vector3f(1, 0, 0)), frame[0])
    assert dr.allclose(to_world.transform_affine(mi.Vector3f(0, 1, 0)), frame[1])
    assert dr.allclose(to_world.transform_affine(mi.Vector3f(0, 0, 1)), frame[2])


def test_distant_flux_viewing_angles(mode_mono):
    # Test default constructor
    d = DistantFluxMeasure(film_resolution=(2, 2))
    expected = (
        np.array(
            [
                [
                    (41.409622, 225),
                    (41.409622, 135),
                ],
                [
                    (41.409622, 315),
                    (41.409622, 45),
                ],
            ]
        )
        * ureg.deg
    )
    assert np.allclose(d.viewing_angles, expected)



import os

import pytest

import eradiate

eradiate.plot.set_style()

# ------------------------------------------------------------------------------
#               Customizable output dir for test artifacts
# ------------------------------------------------------------------------------


def pytest_addoption(parser):
    eradiate_source_dir = os.environ.get("ERADIATE_SOURCE_DIR", ".")
    parser.addoption(
        "--artefact-dir",
        action="store",
        default=os.path.join(eradiate_source_dir, "build/test_artefacts/"),
    )


# See: https://stackoverflow.com/a/55301318/3645374
@pytest.fixture(scope="session")
def artefact_dir(pytestconfig):
    option_value = pytestconfig.getoption("artefact_dir")

    if not os.path.isdir(option_value):
        os.makedirs(option_value)

    return option_value


# ------------------------------------------------------------------------------
#                              Other configuration
# ------------------------------------------------------------------------------


def pytest_configure(config):
    markexpr = config.getoption("markexpr", "False")
    has_slow = "not slow" not in markexpr
    has_regression = "not regression" not in markexpr

    if has_slow:
        print(
            "\033[93m"
            "Running slow tests. To skip them, please run "
            "'pytest -m \"not slow\"' "
            "\033[0m"
        )

    if has_regression:
        print(
            "\033[93m"
            "Running regression tests. To skip them, please run "
            "'pytest -m \"not regression\"' "
            "\033[0m"
        )

    config.addinivalue_line(
        "markers", "slow: marks tests as slow (deselect with -m 'not slow')"
    )

    config.addinivalue_line(
        "markers",
        "regression: marks tests as potentially very slow regression tests "
        "(deselect with -m 'not regression')",
    )


# ------------------------------------------------------------------------------
#                            Pre-process helpers
# ------------------------------------------------------------------------------

# This improves assert handling when using the check_scene_element() function
pytest.register_assert_rewrite("eradiate.test_tools.types.check_scene_element")

# ------------------------------------------------------------------------------
#                                 Mode fixtures
# ------------------------------------------------------------------------------


def generate_fixture(mode):
    @pytest.fixture()
    def fixture():
        import eradiate

        eradiate.set_mode(mode)

    globals()["mode_" + mode] = fixture


for mode in eradiate.modes():
    generate_fixture(mode)
del generate_fixture


def generate_fixture_group(name, modes):
    @pytest.fixture(params=modes)
    def fixture(request):
        mode = request.param
        import eradiate

        eradiate.set_mode(mode)

    globals()["modes_" + name] = fixture


variants = [x for x in eradiate.modes() if x not in {"mono", "ckd"}]  # Remove aliases
variant_groups = {
    "all_mono": [x for x in variants if x.startswith("mono")],
    "all_ckd": [x for x in variants if x.startswith("ckd")],
    "all_mono_ckd": [
        x for x in variants if (x.startswith("mono") or x.startswith("ckd"))
    ],
    "all_single": [x for x in variants if x.endswith("single")],
    "all_double": [x for x in variants if x.endswith("double")],
    "all": variants,
}

for name, variants in variant_groups.items():
    generate_fixture_group(name, variants)
del generate_fixture_group


# ------------------------------------------------------------------------------
#                                 Other fixtures
# ------------------------------------------------------------------------------

from eradiate.test_tools.fixtures import *  # noqa: E402, F401, F403



import pytest

import eradiate
from eradiate.units import unit_registry as ureg


@pytest.fixture(scope="module")
def results_mono():
    # Single-sensor setup, no sample count-split
    eradiate.set_mode("mono_double")
    exp = eradiate.experiments.AtmosphereExperiment(
        atmosphere=None,
        surface={"type": "lambertian", "reflectance": 1.0},
        illumination={"type": "directional", "irradiance": 2.0},
        measures=[
            {
                "type": "hemispherical_distant",
                "film_resolution": (32, 32),
                "spp": 250,
                "srf": {"type": "multi_delta", "wavelengths": 550.0 * ureg.nm},
            }
        ],
    )
    exp.process()
    return exp.measures[0].mi_results, exp


@pytest.fixture(scope="module")
def results_mono_spp():
    # Single-sensor, sample count-split setup
    eradiate.set_mode("mono_double")
    exp = eradiate.experiments.AtmosphereExperiment(
        atmosphere=None,
        surface={"type": "lambertian", "reflectance": 1.0},
        illumination={"type": "directional", "irradiance": 2.0},
        measures=[
            {
                "type": "hemispherical_distant",
                "film_resolution": (32, 32),
                "spp": 250,
                "split_spp": 100,
                "srf": {"type": "multi_delta", "wavelengths": 550.0 * ureg.nm},
            }
        ],
    )
    exp.process()
    return exp.measures[0].results, exp


@pytest.fixture(scope="module")
def results_ckd():
    # Single-sensor setup, no sample count-split, CKD mode
    eradiate.set_mode("ckd_double")
    exp = eradiate.experiments.AtmosphereExperiment(
        atmosphere=None,
        surface={"type": "lambertian", "reflectance": 1.0},
        illumination={"type": "directional", "irradiance": 2.0},
        measures=[
            {
                "type": "hemispherical_distant",
                "film_resolution": (32, 32),
                "spp": 250,
                "srf": {"type": "multi_delta", "wavelengths": 550.0 * ureg.nm},
            }
        ],
    )
    exp.process()
    return exp.measures[0].mi_results, exp



import os.path
import textwrap
from pathlib import Path
from typing import List, Optional

import typer
from rich.console import Console
from ruamel.yaml import YAML
from typing_extensions import Annotated

import eradiate

from ..data._util import get_file_list, known_file_lists
from ..exceptions import DataError

app = typer.Typer()
console = Console(color_system=None)


@app.callback()
def main():
    """
    Manage data.
    """
    pass


@app.command()
def make_registry(
    input_directory: Annotated[
        str,
        typer.Option(
            "--input-directory",
            "-i",
            help="Path to input directory.",
        ),
    ] = ".",
    output_file: Annotated[
        Optional[str],
        typer.Option(
            "--output-file",
            "-o",
            help="Path to output file (default: '<input_directory>/registry.txt').",
        ),
    ] = None,
    rules: Annotated[
        Optional[str],
        typer.Option(
            "--rules",
            "-r",
            help="Path to the registry rule file "
            "(default: '<input_directory>/registry_rules.yml').",
        ),
    ] = None,
    hash_algorithm: Annotated[
        str,
        typer.Option(
            "--hash-algorithm",
            "-a",
            help="Hashing algorithm (default: 'sha256').",
        ),
    ] = "sha256",
):
    """
    Recursively construct a file registry from the current working directory.
    """
    from eradiate.data._core import load_rules, make_registry

    input_directory = Path(input_directory)
    console.print(f"Creating registry file from '{input_directory}'")

    # Load include and exclude rules
    if rules is None:
        rules = input_directory / "registry_rules.yml"
    console.print(f"Using rules in '{rules}'")
    rule_map = load_rules(rules)

    # Write registry
    if output_file is None:
        output_file = input_directory / "registry.txt"
    console.print(f"Writing registry file to '{output_file}'")
    make_registry(
        input_directory,
        output_file,
        includes=rule_map["include"],
        excludes=rule_map["exclude"],
        alg=hash_algorithm,
        show_progress=True,
    )


@app.command()
def update_registries():
    """
    Update local registries for online sources.
    """
    # Update data store registries
    for data_store_id, data_store in eradiate.data.data_store.stores.items():
        if isinstance(data_store, eradiate.data.SafeOnlineDataStore):
            console.print(
                f"[bold cyan]{data_store_id}[/] [{data_store.__class__.__name__}]"
            )
            console.print(f"    Refreshing registry \[{data_store.registry_path}]")
            data_store.registry_delete()
            console.print(
                "    Downloading from "
                f"\[{os.path.join(data_store.base_url, data_store.registry_fname)}]"
            )
            data_store.registry_fetch()
            console.print("  [green]✓ Done[/]")
            continue


@app.command()
def fetch(
    file_list: Annotated[
        Optional[List[str]],
        typer.Argument(
            help="An arbitrary number of relative paths to files to be "
            "retrieved from the data store or file list specifications. "
            "If unset, the list of files is read from a YAML file which can be "
            "specified by using the ``--from-file`` option, if it is specified. "
            "Otherwise, it defaults to ``all`` in a production environment and "
            "``minimal`` in a development environment."
        ),
    ] = None,
    from_file: Annotated[
        Optional[str],
        typer.Option(
            "--from-file",
            "-f",
            help="Optional path to a file list (YAML format). If this option "
            "is set, the FILES argument(s) will be ignored.",
        ),
    ] = None,
    list: Annotated[
        bool,
        typer.Option(
            "--list",
            "-l",
            help="List built-in file lists and exit.",
        ),
    ] = False,
):
    """
    Fetch files from the Eradiate data store.
    """
    if list:
        print("Known file lists:")
        for file_list in known_file_lists():
            print(f"  {file_list}")
        exit(0)

    if not file_list:
        if from_file is None:
            if eradiate.config.SOURCE_DIR is None:
                file_list = ["all"]
            else:
                file_list = ["minimal"]
        else:
            console.print(f"Reading file list from '{from_file}'")
            yaml = YAML()
            file_list = yaml.load(Path(from_file))

    converted = []
    for spec in file_list:
        try:
            converted.extend(get_file_list(spec))
        except ValueError:
            converted.append(spec)

    file_list = converted

    for filename in file_list:
        try:
            console.print(f"[blue]Fetching '{filename}'[/]")
            path = eradiate.data.data_store.fetch(filename)
        except DataError:
            console.print("[red]✗[/] not found")
        else:
            console.print(f"[green]✓[/] found \[{path}]")


@app.command()
def purge_cache(
    keep: Annotated[
        bool, typer.Option("--keep", "-k", help="Keep registered files.")
    ] = False,
):
    """
    Purge the cache of online data stores.
    """
    for data_store_id, data_store in eradiate.data.data_store.stores.items():
        console.print(
            f"[bold cyan]{data_store_id}[/] [{data_store.__class__.__name__}]"
        )

        if isinstance(data_store, eradiate.data.SafeDirectoryDataStore):
            console.print("  Skipping")
            continue

        if isinstance(data_store, eradiate.data.SafeOnlineDataStore):
            console.print(f"  Purging '{data_store.path}'")
            if keep:
                data_store.purge(keep="registered")
            else:
                data_store.purge()
            continue

        if isinstance(data_store, eradiate.data.BlindOnlineDataStore):
            console.print(f"  Purging '{data_store.path}'")
            data_store.purge()
            continue


@app.command()
def info(
    data_stores: Annotated[
        Optional[List[str]],
        typer.Argument(
            help=(
                "List of data stores for which information is requested. "
                "If no data store ID is passed, information is displayed for "
                "all data stores."
            )
        ),
    ] = None,
    list_registry: Annotated[
        bool,
        typer.Option(
            "-l", "--list-registry", help="Show registry content if relevant."
        ),
    ] = False,
):
    """
    Display information about data store configuration.
    """

    # Build section list
    sections = [
        ("[purple]Base URL[/] \[base_url]", "base_url"),
        ("[purple]Local path[/] \[path]", "path"),
        ("[purple]Registry path[/] \[registry_path]", "registry_path"),
    ]

    if list_registry:
        sections.append(("Registered files", "registry_keys"))

    first = True

    for data_store_id, data_store in eradiate.data.data_store.stores.items():
        if data_stores and (data_store_id not in data_stores):
            continue

        # Collect section contents
        reprs = {}

        for attr in ["base_url", "path", "registry_path"]:
            try:
                reprs[attr] = f"'{getattr(data_store, attr)}'"
            except (NotImplementedError, AttributeError):
                reprs[attr] = None

        try:
            reprs["registry_keys"] = "\n".join(
                f"'{x}'" for x in sorted(data_store.registry.keys())
            )
        except (NotImplementedError, AttributeError):
            reprs["registry_keys"] = None

        # Display the content for current data store
        if not first:
            console.print()
        else:
            first = False

        console.print(
            f"[bold cyan]{data_store_id}[/] \[{data_store.__class__.__name__}]"
        )

        for section_title, repr_key in sections:
            the_repr = reprs[repr_key]

            if the_repr is not None:
                console.print(f"  {section_title}")
                console.print(textwrap.indent(reprs[repr_key], " " * 4))



import mitsuba as mi
import numpy as np
import pytest

from eradiate import KernelContext
from eradiate import unit_registry as ureg
from eradiate.kernel import mi_render, mi_traverse
from eradiate.scenes.biosphere import LeafCloud
from eradiate.scenes.core import Scene, traverse
from eradiate.scenes.illumination import DirectionalIllumination
from eradiate.scenes.integrators import PathIntegrator
from eradiate.scenes.measure import PerspectiveCameraMeasure
from eradiate.scenes.surface import BasicSurface

wavelengths = np.linspace(500, 600, 21) * ureg.nm
spp = 100


def make_scene():
    return Scene(
        objects={
            "surface": BasicSurface(shape={"type": "rectangle"}),
            "canopy": LeafCloud.cuboid(
                n_leaves=10000,
                leaf_radius=5 * ureg.cm,
                l_horizontal=10 * ureg.m,
                l_vertical=1 * ureg.m,
            ),
            "illumination": DirectionalIllumination(),
            "integrator": PathIntegrator(),
            "measure": PerspectiveCameraMeasure(
                target=[0, 0, 0], origin=[2, 2, 2], up=[0, 0, 1]
            ),
        }
    )


def test_mi_render(mode_mono):
    """
    Mitsuba scene render
    ====================

    This basic test checks that the ``mi_render()`` function runs without
    failing and stores results as expected.

    Rationale
    ---------

    We construct a simple scene and render it using our ``mi_render()`` wrapper
    for a sequence of wavelengths.

    Expected behaviour
    ------------------

    Rendering succeed and results are stored in a dictionary which uses
    parametric loop index values (in this case, the wavelength) as keys.
    """
    template, params = traverse(make_scene())
    mi_scene = mi_traverse(mi.load_dict(template.render(ctx=KernelContext())))
    contexts = [KernelContext(si={"w": w}) for w in wavelengths]
    result = mi_render(mi_scene, contexts, spp=spp)

    # We store film values and SPPs
    assert set(result.keys()) == set(x.si.w.magnitude for x in contexts)
    assert set(result[500.0].keys()) == {"measure"}


@pytest.mark.slow
def test_mi_render_rebuild(mode_mono):
    """
    Mitsuba render scene rebuild
    ============================

    This test performs the same computation as the previous (see
    *Mitsuba scene render*), but rebuilds the scene for each parametric
    iteration. It is designed for comparison with the previous.

    Rationale
    ---------

    The *Mitsuba scene render* test does not reload the scene at each iteration;
    instead, it uses Mitsuba's parameter update system to update it at a low
    cost. In this test, the scene is entirely rebuilt at each iteration.

    Expected behaviour
    ------------------

    The rendering sequence succeeds and yields the same results as the
    *Mitsuba scene render* test. This test should take much longer than the
    *Mitsuba render scene* to complete.
    """
    # TODO: Recycle this test and merge with previous
    template, params = traverse(make_scene())
    kdict = template.render(ctx=KernelContext(), drop=True)

    for w in wavelengths:
        mi_scene = mi_traverse(mi.load_dict(kdict))
        mi_render(
            mi_scene,
            [KernelContext(si={"w": w})],
            spp=spp,
        )



from eradiate.data._core import registry_from_file, registry_to_file


def test_registry_from_file(tmpdir):
    filename = tmpdir / "registry.txt"
    with open(filename, "w") as f:
        f.write(
            """
            # Registry test file
            foo1.nc
            foo2.nc sha256:bar
            foo3.nc sha256:bar https://www.foo.bar
            # Another comment line

            # Empty line above
            # By the way, the indentation will be stripped upon reading
            """
        )

    registry = registry_from_file(filename)
    assert registry == {
        "foo1.nc": "",
        "foo2.nc": "sha256:bar",
        "foo3.nc": "sha256:bar https://www.foo.bar",
    }


def test_registry_to_file(tmpdir):
    filename = tmpdir / "registry.txt"
    registry = {
        "foo1.nc": "",
        "foo2.nc": "sha256:bar",
        "foo3.nc": "sha256:bar https://www.foo.bar",
    }

    registry_to_file(registry, filename)

    with open(filename, "r") as f:
        s = f.read()

    assert s == "foo1.nc\nfoo2.nc sha256:bar\nfoo3.nc sha256:bar https://www.foo.bar\n"



import mitsuba as mi
import numpy as np
import pytest
import xarray as xr

import eradiate
from eradiate import unit_registry as ureg
from eradiate.constants import EARTH_RADIUS
from eradiate.experiments import DEMExperiment
from eradiate.scenes.atmosphere import HomogeneousAtmosphere
from eradiate.scenes.geometry import PlaneParallelGeometry
from eradiate.scenes.measure import MultiDistantMeasure
from eradiate.scenes.spectra import MultiDeltaSpectrum
from eradiate.scenes.surface import DEMSurface, mesh_from_dem
from eradiate.test_tools.types import check_scene_element


def test_dem_experiment_construct_default(modes_all_double):
    """
    DEMExperiment initialises with default params in all modes
    """
    exp = DEMExperiment()

    # Check that the atmosphere's geometry is overridden by the experiment's
    assert exp.atmosphere.geometry is exp.geometry


def test_dem_experiment_construct_measures(modes_all):
    """
    A variety of measure specifications are acceptable
    """
    # Init with a single measure (not wrapped in a sequence)
    assert DEMExperiment(measures=MultiDistantMeasure())

    # Init from a dict-based measure spec
    # -- Correctly wrapped in a sequence
    assert DEMExperiment(measures=[{"type": "distant"}])

    # -- Not wrapped in a sequence
    assert DEMExperiment(measures={"type": "distant"})


def test_dem_experiment_construct_normalize_measures(mode_mono):
    # When setting atmosphere to None, measure target is at ground level
    exp = DEMExperiment(atmosphere=None)
    assert np.allclose(exp.measures[0].target.xyz, [0, 0, 0] * ureg.m)

    # When atmosphere is set, measure target is at ground level
    exp = DEMExperiment(atmosphere=HomogeneousAtmosphere())
    assert np.allclose(exp.measures[0].target.xyz, [0, 0, 0] * ureg.m)


def test_dem_experiment_ckd(mode_ckd, atmosphere_us_standard_ckd):
    """
    DEMExperiment with heterogeneous atmosphere in CKD mode can be created.
    """
    exp = DEMExperiment(
        atmosphere=atmosphere_us_standard_ckd,
        surface={"type": "lambertian"},
        measures={"type": "distant", "id": "distant_measure"},
    )
    check_scene_element(exp.scene, mi.Scene, ctx=exp.context_init)


def test_dem_experiment_kernel_dict(modes_all_double):
    """
    Test non-trivial kernel dict generation behaviour.
    """

    # Setting atmosphere to None
    exp = DEMExperiment(
        atmosphere=None,
        surface={"type": "lambertian"},
        measures=[
            {"type": "distant", "id": "distant_measure"},
            {"type": "radiancemeter", "origin": [1, 0, 0], "id": "radiancemeter"},
        ],
    )
    # -- Surface has default value
    mi_wrapper = check_scene_element(
        exp.scene, mi.Scene, ctx=exp.context_init, drop_parameters=False
    )  # Do not drop untracked parameters: we want to check the surface transform
    np.testing.assert_allclose(
        mi_wrapper.parameters["surface_shape.to_world"].matrix,
        mi.ScalarTransform4f.scale([5e8, 5e8, 1]).matrix,
    )
    # -- Atmosphere is not in kernel dictionary
    assert {shape.id() for shape in mi_wrapper.obj.shapes()} == {"surface_shape"}

    # -- Measures get no external medium assigned
    assert all(sensor.medium() is None for sensor in mi_wrapper.obj.sensors())


@pytest.mark.slow
def test_dem_experiment_real_life(mode_mono, atmosphere_us_standard_mono):
    # Construct with typical parameters
    exp = DEMExperiment(
        surface={"type": "rpv"},
        atmosphere={
            "type": "heterogeneous",
            "molecular_atmosphere": atmosphere_us_standard_mono,
        },
        illumination={"type": "directional", "zenith": 45.0},
        measures=[
            {"type": "distant", "id": "distant_measure"},
            {"type": "radiancemeter", "origin": [1, 0, 0], "id": "radiancemeter"},
        ],
    )

    mi_wrapper = check_scene_element(exp.scene, mi.Scene, ctx=exp.context_init)

    # -- Distant measures get no external medium
    assert mi_wrapper.obj.sensors()[0].medium() is None

    # -- Radiancemeter inside the atmosphere must have a medium assigned
    assert mi_wrapper.obj.sensors()[1].medium().id() == "medium_atmosphere"


def test_dem_experiment_inconsistent_multiradiancemeter(
    mode_mono,
    atmosphere_us_standard_mono,
):
    # A MultiRadiancemeter measure must have all origins inside the atmosphere or none.
    # A mix of both will raise an error.

    # Construct with typical parameters
    exp = DEMExperiment(
        surface={"type": "rpv"},
        atmosphere={
            "type": "heterogeneous",
            "molecular_atmosphere": atmosphere_us_standard_mono,
        },
        illumination={"type": "directional", "zenith": 45.0},
        measures=[
            {
                "type": "multi_radiancemeter",
                "origins": [[0, 0, 1], [0, 0, 1000000]],
                "directions": [[0, 0, -1], [0, 0, -1]],
                "id": "multi_radiancemeter",
            },
        ],
    )

    with pytest.raises(
        ValueError,
        match="Inconsistent placement of MultiRadiancemeterMeasure origins",
    ):
        exp.init()


def test_dem_experiment_run_basic(modes_all_double):
    """
    DEMExperiment runs successfully in all modes.
    """
    exp = DEMExperiment()
    exp.measures[0].srf = MultiDeltaSpectrum(wavelengths=550.0 * ureg.nm)
    eradiate.run(exp)
    assert isinstance(exp.results, dict)


@pytest.mark.slow
def test_dem_experiment_run_detailed(modes_all_double):
    """
    Test for correctness of the result dataset generated by DEMExperiment.
    """
    # Create simple scene
    exp = DEMExperiment(
        measures=[
            {
                "type": "hemispherical_distant",
                "id": "toa_hsphere",
                "film_resolution": (32, 32),
                "spp": 1000,
                "srf": MultiDeltaSpectrum(wavelengths=550.0 * ureg.nm),
            },
        ]
    )

    # Run RT simulation
    results = eradiate.run(exp)

    # Check result dataset structure
    # Post-processing creates expected variables ...
    expected = {"brdf", "brf", "irradiance", "radiance"}
    assert set(results.data_vars) == expected

    # ... dimensions
    assert set(results["radiance"].dims) == {"sza", "saa", "x_index", "y_index", "w"}
    assert set(results["irradiance"].dims) == {"sza", "saa", "w"}

    # ... and other coordinates
    expected_coords = {"sza", "saa", "vza", "vaa", "x", "y", "x_index", "y_index", "w"}
    if eradiate.mode().is_ckd:
        expected_coords |= {"bin_wmin", "bin_wmax"}
    assert set(results["radiance"].coords) == expected_coords

    expected_coords = {"sza", "saa", "w"}
    if eradiate.mode().is_ckd:
        expected_coords |= {"bin_wmin", "bin_wmax"}
    assert set(results["irradiance"].coords) == expected_coords

    # We just check that we record something as expected
    assert np.all(results["radiance"].data > 0.0)


def test_dem_experiment_warn_targeting_dem(modes_all_double, tmpdir):
    """
    Test that the DEMExperiment constructor raises a warning when the measure
    target is a point and a DEM is defined in the scene.
    """

    da = xr.DataArray(
        data=np.zeros((10, 10)),
        dims=["x", "y"],
        coords={
            "x": (["x"], np.linspace(-10, 20, 10), {"units": "kilometer"}),
            "y": (["y"], np.linspace(-30, 40, 10), {"units": "kilometer"}),
        },
        attrs={"units": "meter"},
    )

    mesh, lat, lon = mesh_from_dem(
        da, geometry="plane_parallel", planet_radius=EARTH_RADIUS
    )

    with pytest.warns(UserWarning, match="uses a point target"):
        DEMExperiment(
            surface=DEMSurface.from_mesh(
                id="terrain",
                mesh=mesh,
                lat=lat,
                lon=lon,
                geometry=PlaneParallelGeometry(),
            ),
            measures=[
                {
                    "type": "distant",
                    "id": "distant_measure",
                    "target": {"type": "point", "xyz": [1, 1, 0]},
                },
            ],
        )



import drjit as dr
import mitsuba as mi
import numpy as np
import pytest

from eradiate import KernelContext
from eradiate.exceptions import TraversalError
from eradiate.scenes.core import Ref, Scene, traverse
from eradiate.scenes.shapes import RectangleShape, Shape
from eradiate.scenes.surface import CentralPatchSurface
from eradiate.units import unit_registry as ureg


@pytest.mark.parametrize(
    "kwargs, expected_shape, expected_traversal_param_keys",
    [
        ({}, None, TraversalError),
        (
            {"shape": {"type": "rectangle"}},
            RectangleShape,
            {"surface_bsdf.bsdf_0.reflectance.value"},
        ),
        (
            {"shape": {"type": "rectangle"}, "patch_edges": 1.0},
            RectangleShape,
            {"surface_bsdf.bsdf_0.reflectance.value"},
        ),
        (
            {"shape": {"type": "rectangle"}, "patch_edges": [1.0, 0.5]},
            RectangleShape,
            {"surface_bsdf.bsdf_0.reflectance.value"},
        ),
    ],
    ids=[
        "noargs",
        "shape",
        "edges_scalar",
        "edges_vector",
    ],
)
def test_central_patch_construct_new(
    modes_all_double, kwargs, expected_shape, expected_traversal_param_keys
):
    surface = CentralPatchSurface(**kwargs)

    if expected_shape is None:
        assert surface.shape is None
    elif issubclass(expected_shape, Shape):
        assert isinstance(surface.shape, expected_shape)
        # Shape BSDF is referenced
        assert isinstance(surface.shape.bsdf, Ref)
    else:
        raise NotImplementedError

    if isinstance(expected_traversal_param_keys, set):
        template, params = traverse(surface)

        # Scene element is composite: template has not "type" key
        assert "type" not in template
        # Parameter map keys are fetched recursively
        assert set(params.keys()) == expected_traversal_param_keys

        # When enclosed in a Scene, the surface can be traversed
        scene = Scene(objects={"surface": surface})
        template, params = traverse(scene)
        kernel_dict = template.render(KernelContext())
        assert isinstance(mi.load_dict(kernel_dict), mi.Scene)

    elif isinstance(expected_traversal_param_keys, type) and issubclass(
        expected_traversal_param_keys, TraversalError
    ):
        with pytest.raises(
            expected_traversal_param_keys,
            match="A 'CentralPatchSurface' cannot be traversed if its 'shape' field is unset",
        ):
            traverse(surface)
    else:
        raise NotImplementedError


def test_central_patch_texture_scale(mode_mono):
    # Default value: the patch is not scaled
    surface = CentralPatchSurface(shape={"type": "rectangle", "edges": 10.0})
    assert np.allclose(1.0, surface._texture_scale())

    # Specify edge values
    surface = CentralPatchSurface(
        shape={"type": "rectangle", "edges": 10.0},
        patch_edges=[10.0, 10.0 / 3.0],
    )
    assert np.allclose(surface._texture_scale(), [1.0 / 3.0, 1.0])


def test_central_patch_scale_kernel_dict(mode_mono):
    surface = CentralPatchSurface(
        shape={"type": "rectangle", "edges": 3000.0 * ureg.km},
        patch_edges=100 * ureg.km,
        id="surface",
    )

    template, params = traverse(surface)
    kernel_dict = template.render(ctx=KernelContext())
    result = kernel_dict["surface_bsdf"]["weight"]["to_uv"].matrix
    expected = (
        mi.ScalarTransform4f.scale([10, 10, 1])
        @ mi.ScalarTransform4f.translate((-0.45, -0.45, 0))
    ).matrix

    assert dr.allclose(result, expected)



import numpy as np
import pytest

from eradiate import unit_registry as ureg
from eradiate.quad import Quad, QuadType
from eradiate.spectral import Bin, BinSet
from eradiate.spectral.ckd import (
    QuadSpec,
    QuadSpecErrorThreshold,
    QuadSpecFixed,
    QuadSpecMinError,
)


def test_ckd_bin():
    """
    Unit tests for :class:`eradiate.ckd.Bin`.
    """
    quad = Quad.gauss_legendre(16)

    # Construct a bin
    bin = Bin(wmin=500.0, wmax=510.0, quad=quad)

    # Scalar values are correctly converted to config default units
    assert bin.wmin == 500.0 * ureg.nm

    # Wrong bound ordering raises
    with pytest.raises(ValueError):
        Bin(wmin=510.0, wmax=500.0, quad=quad)


def test_bin_set_arange():
    binset = BinSet.arange(
        500 * ureg.nm, 600 * ureg.nm, 10 * ureg.nm, quad=Quad.gauss_legendre(2)
    )
    result = binset.wavelengths
    expected = [505, 515, 525, 535, 545, 555, 565, 575, 585, 595] * ureg.nm
    assert np.allclose(result, expected)


def test_quad_spec_default():
    assert QuadSpec.default() == QuadSpecFixed(quad_type=QuadType.GAUSS_LEGENDRE, n=16)


@pytest.mark.parametrize(
    "value, expected",
    [
        (
            {"type": "fixed", "n": 8, "quad_type": "gauss_lobatto"},
            QuadSpecFixed(8, "gauss_lobatto"),
        ),
        (
            {"type": "minimize", "nmax": 8},
            QuadSpecMinError(8),
        ),
        (
            {"type": "threshold", "nmax": 8, "threshold": 0.01},
            QuadSpecErrorThreshold(0.01, 8),
        ),
        ({"type": "invalid"}, ValueError),
        ({}, ValueError),
    ],
)
def test_quad_spec_from_dict(value, expected):
    """
    Unit tests for :class:`.QuadSpec`.
    """
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            QuadSpec.from_dict(value)

    else:
        quad_spec = QuadSpec.from_dict(value)
        assert quad_spec == expected



"""
Tests for the radprops._absorption module.
"""

import numpy as np
import pytest

import eradiate
from eradiate import unit_registry as ureg
from eradiate.radprops import (
    CKDAbsorptionDatabase,
    ErrorHandlingConfiguration,
    MonoAbsorptionDatabase,
)

# ------------------------------------------------------------------------------
#                                    Fixtures
# ------------------------------------------------------------------------------


@pytest.mark.usefixtures("mode_mono")
class TestMonoAbsorptionDatabase:
    @pytest.fixture(scope="class")
    def db(self):
        eradiate.set_mode("none")
        _db = MonoAbsorptionDatabase.from_name("komodo")
        yield _db
        _db.cache_clear()

    def test_construct(self, db):
        # Default komodo settings use lazy data loading
        assert db.lazy is True

        # The dict converter accepts kwargs and can be used to override defaults
        db = MonoAbsorptionDatabase.from_dict(
            {"construct": "from_name", "name": "komodo", "lazy": False}
        )
        assert db.lazy is False

    @pytest.mark.parametrize(
        "w",
        [
            np.array([550.0]) * ureg.nm,
            np.linspace(540.0, 560.0) * ureg.nm,
        ],
    )
    def test_eval(
        self, db, thermoprops_us_standard, absorption_database_error_handler_config, w
    ):
        sigma_a = db.eval_sigma_a_mono(
            w,
            thermoprops_us_standard,
            ErrorHandlingConfiguration.convert(
                absorption_database_error_handler_config
            ),
        )

        # sigma_a should have a shape of (w, z)
        z = thermoprops_us_standard.z.values
        assert sigma_a.values.shape == (w.size, z.size)


@pytest.mark.usefixtures("mode_ckd")
class TestCKDAbsorptionDatabase:
    @pytest.fixture(scope="class")
    def db(self):
        eradiate.set_mode("none")
        _db = CKDAbsorptionDatabase.from_name("monotropa")
        yield _db
        _db.cache_clear()

    def test_construct(self, db):
        # Default monotropa settings use eager data loading
        assert db.lazy is False

        # The dict converter accepts kwargs and can be used to override defaults
        db = MonoAbsorptionDatabase.from_dict(
            {"construct": "from_name", "name": "monotropa", "lazy": True}
        )
        assert db.lazy is True

    @pytest.mark.parametrize(
        "w, expected",
        [
            ({"wl": 550.0}, ["monotropa-18100_18200.nc"]),
            ({"wl": 550.0 * ureg.nm}, ["monotropa-18100_18200.nc"]),
            ({"wl": 0.55 * ureg.micron}, ["monotropa-18100_18200.nc"]),
            ({"wl": [550.0, 550.0]}, ["monotropa-18100_18200.nc"] * 2),
        ],
        ids=[
            "wl_scalar_unitless",
            "wl_scalar_nm",
            "wl_scalar_micron",
            "wl_array_unitless",
        ],
    )
    def test_filename_lookup(self, db, w, expected):
        assert db.lookup_filenames(**w) == expected

    @pytest.mark.parametrize(
        "wg",
        [([550.0] * ureg.nm, 0.5)],
    )
    def test_eval(
        self, db, thermoprops_us_standard, absorption_database_error_handler_config, wg
    ):
        sigma_a = db.eval_sigma_a_ckd(
            *wg,
            thermoprops_us_standard,
            ErrorHandlingConfiguration.convert(
                absorption_database_error_handler_config
            ),
        )

        # sigma_a should have a shape of (w, z)
        z = thermoprops_us_standard.z.values
        assert sigma_a.values.shape == (wg[0].size, z.size)

    def test_cache_clear(self, db):
        # Make a query to ensure that the cache is filling up
        db.load_dataset("monotropa-18100_18200.nc")
        assert db._cache.currsize > 0
        # Clear the cache: it should be empty after that
        db.cache_clear()
        assert db._cache.currsize == 0

    def test_cache_reset(self, db):
        db.cache_reset(2)
        assert db._cache.currsize == 0
        assert db._cache.maxsize == 2
        db.cache_reset(8)
        assert db._cache.currsize == 0
        assert db._cache.maxsize == 8



from pathlib import Path

import pytest

import eradiate
from eradiate.data import BlindDirectoryDataStore
from eradiate.exceptions import DataError

TEST_FILE_EXISTS = Path("tests/data/git/registered_dataset.nc")
TEST_FILE_DOES_NOT_EXIST = Path("tests/data/git/does_not_exist.nc")


def test_directory_data_store_fetch():
    # The data submodule can be instantiated
    data_store = BlindDirectoryDataStore(
        path=eradiate.config.SOURCE_DIR / "resources/data"
    )

    # We can fetch the test file
    assert data_store.fetch(TEST_FILE_EXISTS)

    # A missing file raises
    with pytest.raises(DataError):
        data_store.fetch(TEST_FILE_DOES_NOT_EXIST)



import mitsuba as mi
import numpy as np
import pytest

from eradiate import unit_registry as ureg
from eradiate.scenes.measure import HemisphericalDistantMeasure
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "tested",
    [
        {},
        dict(target=[0, 0, 0]),
        dict(target={"type": "rectangle", "xmin": 0, "xmax": 1, "ymin": 0, "ymax": 1}),
    ],
    ids=[
        "no_args",
        "target_point",
        "target_rectangle",
    ],
)
def test_hemispherical_distant_construct(modes_all_double, tested):
    measure = HemisphericalDistantMeasure(**tested)
    check_scene_element(measure, mi.Sensor)


def test_hemispherical_distant_viewing_angles(mode_mono):
    # Test default constructor
    d = HemisphericalDistantMeasure(film_resolution=(2, 2))
    expected = (
        np.array(
            [
                [
                    (41.409622, 225),
                    (41.409622, 135),
                ],
                [
                    (41.409622, 315),
                    (41.409622, 45),
                ],
            ]
        )
        * ureg.deg
    )
    assert np.allclose(d.viewing_angles, expected)



import os
import tempfile

import numpy as np
import pytest

from eradiate import unit_registry as ureg
from eradiate.scenes.biosphere import (
    AbstractTree,
    DiscreteCanopy,
    InstancedCanopyElement,
    LeafCloud,
    MeshTree,
    MeshTreeElement,
)
from eradiate.scenes.core import traverse
from eradiate.test_tools.types import check_scene_element

# ------------------------------------------------------------------------------
#                            Fixture definitions
# ------------------------------------------------------------------------------


@pytest.fixture(scope="module")
def tempfile_leaves():
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, "tempfile_leaves.txt")
        with open(filename, "w") as tf:
            tf.write("0.100 8.864 9.040 1.878 -0.314 0.025 0.949\n")
            tf.write("0.100 9.539 -10.463 0.627 0.489 -0.276 0.828\n")
            tf.write("0.100 -2.274 -9.204 0.797 0.618 0.184 0.764\n")
            tf.write("0.100 -9.957 -4.971 0.719 -0.066 0.100 0.993\n")
            tf.write("0.100 5.339 9.153 0.500 0.073 -0.294 0.953\n")
        yield filename


@pytest.fixture(scope="module")
def tempfile_spheres():
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, "tempfile_leaves.txt")
        with open(filename, "w") as tf:
            tf.write("8.864 9.040 1.878\n")
            tf.write("0.100 9.539 -10.463\n")
            tf.write("0.100 -2.274 -9.204\n")
            tf.write("0.100 -9.957 -4.971\n")
            tf.write("0.100 5.339 9.153\n")
        yield filename


@pytest.fixture(scope="module")
def tempfile_obj():
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, "tempfile_mesh.obj")
        with open(filename, "w") as tf:
            tf.write(
                """o Cube
v 1.000000 -1.000000 -1.000000
v 1.000000 -1.000000 1.000000
v -1.000000 -1.000000 1.000000
v -1.000000 -1.000000 -1.000000
v 1.000000 1.000000 -0.999999
v 0.999999 1.000000 1.000001
v -1.000000 1.000000 1.000000
v -1.000000 1.000000 -1.000000
vt 1.000000 0.333333
vt 1.000000 0.666667
vt 0.666667 0.666667
vt 0.666667 0.333333
vt 0.666667 0.000000
vt 0.000000 0.333333
vt 0.000000 0.000000
vt 0.333333 0.000000
vt 0.333333 1.000000
vt 0.000000 1.000000
vt 0.000000 0.666667
vt 0.333333 0.333333
vt 0.333333 0.666667
vt 1.000000 0.000000
vn 0.000000 -1.000000 0.000000
vn 0.000000 1.000000 0.000000
vn 1.000000 0.000000 0.000000
vn -0.000000 0.000000 1.000000
vn -1.000000 -0.000000 -0.000000
vn 0.000000 0.000000 -1.000000
s off
f 2/1/1 3/2/1 4/3/1
f 8/1/2 7/4/2 6/5/2
f 5/6/3 6/7/3 2/8/3
f 6/8/4 7/5/4 3/4/4
f 3/9/5 7/10/5 8/11/5
f 1/12/6 4/13/6 8/11/6
f 1/4/1 2/1/1 4/3/1
f 5/14/2 8/1/2 6/5/2
f 1/12/3 5/6/3 2/8/3
f 2/12/4 6/8/4 3/4/4
f 4/13/5 3/9/5 8/11/5
f 5/6/6 1/12/6 8/11/6"""
            )
        yield filename


# ------------------------------------------------------------------------------
#                             DiscreteCanopy Tests
# ------------------------------------------------------------------------------


def test_discrete_canopy_instantiate(mode_mono):
    # Constructing an empty canopy is possible
    assert DiscreteCanopy()


def test_discrete_canopy_homogeneous(mode_mono):
    # The generate_homogeneous() constructor returns a valid canopy object
    canopy = DiscreteCanopy.homogeneous(
        n_leaves=1, leaf_radius=0.1, l_horizontal=10, l_vertical=3
    )
    check_scene_element(canopy)

    # Check template and parameter map contents
    template, params = traverse(canopy)


def test_discrete_canopy_from_files(mode_mono, tempfile_spheres, tempfile_leaves):
    # The from_files() constructor returns a valid canopy object
    canopy = DiscreteCanopy.leaf_cloud_from_files(
        size=[1, 1, 1],
        leaf_cloud_dicts=[
            {
                "sub_id": "spheres_1",
                "instance_filename": tempfile_spheres,
                "leaf_cloud_filename": tempfile_leaves,
            },
            {
                "sub_id": "spheres_2",
                "instance_filename": tempfile_spheres,
                "leaf_cloud_filename": tempfile_leaves,
            },
        ],
    )
    check_scene_element(canopy)


def test_discrete_canopy_advanced(
    mode_mono, tempfile_spheres, tempfile_leaves, tempfile_obj
):
    """
    A more advanced test where we load a series of different canopy elements:

    - A pre-computed canopy consisting of a generated cuboid leaf cloud
    - A series of instanced leaf clouds from files
    - An abstract tree with a leaf cloud
    - A mesh based canopy element
    """

    canopy = DiscreteCanopy(
        size=[1.0, 1.0, 1.0] * ureg.m,
        instanced_canopy_elements=[
            InstancedCanopyElement(
                instance_positions=[[0, 0, 0]],
                canopy_element=LeafCloud.cuboid(
                    n_leaves=1,
                    leaf_radius=0.1,
                    l_horizontal=10,
                    l_vertical=3,
                    id="leaf_cloud_cuboid",
                ),
            ),
            InstancedCanopyElement.from_file(
                filename=tempfile_spheres,
                canopy_element=LeafCloud.from_file(
                    filename=tempfile_leaves,
                    id="leaf_cloud_precomputed",
                ),
            ),
            InstancedCanopyElement(
                instance_positions=[[0, 0, 0]],
                canopy_element=AbstractTree(
                    leaf_cloud=LeafCloud.cuboid(
                        n_leaves=100,
                        l_horizontal=10.0,
                        l_vertical=1.0,
                        leaf_radius=10.0 * ureg.cm,
                    ),
                    trunk_height=2.0,
                    trunk_radius=0.2,
                    trunk_reflectance=0.5,
                    id="abstract_tree",
                ),
            ),
            InstancedCanopyElement(
                instance_positions=[[0, 0, 0]],
                canopy_element=MeshTree(
                    id="mesh_tree",
                    mesh_tree_elements=[
                        MeshTreeElement(
                            mesh_filename=tempfile_obj,
                            mesh_units=ureg.m,
                            reflectance=0.5,
                            transmittance=0.5,
                            id="mesh_element",
                        )
                    ],
                ),
            ),
        ],
    )
    check_scene_element(canopy)


def test_discrete_canopy_padded(mode_mono, tempfile_leaves, tempfile_spheres):
    """Unit tests for :meth:`.DiscreteCanopy.padded`"""

    canopy = DiscreteCanopy.leaf_cloud_from_files(
        id="canopy",
        size=[100, 100, 30],
        leaf_cloud_dicts=[
            {
                "instance_filename": tempfile_spheres,
                "leaf_cloud_filename": tempfile_leaves,
                "sub_id": "leaf_cloud",
            }
        ],
    )
    # The padded canopy object is valid and instantiable
    padded_canopy = canopy.padded_copy(2)
    assert isinstance(padded_canopy, DiscreteCanopy)
    assert check_scene_element(padded_canopy)
    # Padded canopy has (2*padding + 1) ** 2 times more instances than original
    assert (
        len(padded_canopy._template_instances) == len(canopy._template_instances) * 25
    )
    # Padded canopy has 2*padding + 1 times larger horizontal size than original
    assert np.allclose(padded_canopy.size[:2], 5 * canopy.size[:2])
    # Padded canopy has same vertical size as original
    assert padded_canopy.size[2] == canopy.size[2]



import mitsuba as mi
import numpy as np
import pytest
import xarray as xr

import eradiate
from eradiate.scenes.phase import TabulatedPhaseFunction
from eradiate.spectral import SpectralIndex
from eradiate.test_tools.types import check_scene_element


@pytest.fixture
def regular() -> xr.DataArray:
    """Phase function data with a regular mu grid."""
    result = eradiate.data.load_dataset("tests/spectra/particles/random-regular_mu.nc")
    return result.phase


@pytest.fixture
def irregular() -> xr.DataArray:
    """Phase function data with an irregular mu grid."""
    result = eradiate.data.load_dataset(
        "tests/spectra/particles/random-irregular_mu.nc"
    )
    return result.phase


@pytest.mark.parametrize("grid", ["regular", "irregular"])
def test_tabulated_construct(modes_all_double, grid, request):
    phase = TabulatedPhaseFunction(data=request.getfixturevalue(grid))

    # Irregular grid layout is detected
    assert phase._is_irregular == (grid == "irregular")
    assert (
        phase.kernel_type == "tabphase" if "grid" == "regular" else "tabphase_irregular"
    )

    # Kernel dict can be generated and instantiated
    check_scene_element(phase, mi.PhaseFunction)


def test_tabulated_order(mode_mono, tmpdir):
    """
    TabulatedPhaseFunction returns phase function values by increasing order of
    scattering angle cosine values, regardless how its input is ordered.
    """

    def make_da(mu, phase):
        return xr.DataArray(
            phase[:, :, np.newaxis, np.newaxis],
            coords={
                "w": ("w", [240.0, 2800.0], dict(units="nm")),
                "mu": ("mu", mu),
                "i": ("i", [0]),
                "j": ("j", [0]),
            },
        )

    da_mu_increasing = make_da(
        mu=np.linspace(-1, 1, 3), phase=np.array([np.arange(1, 4), np.arange(1, 4)])
    )

    da_mu_decreasing = make_da(
        mu=np.linspace(1, -1, 3),
        phase=np.array([np.arange(3, 0, -1), np.arange(3, 0, -1)]),
    )

    si = SpectralIndex.new()

    layer_mu_increasing = TabulatedPhaseFunction(data=da_mu_increasing)
    phase_mu_increasing = layer_mu_increasing.eval(si)

    layer_mu_decreasing = TabulatedPhaseFunction(data=da_mu_decreasing)
    phase_mu_decreasing = layer_mu_decreasing.eval(si)

    assert np.all(phase_mu_increasing == phase_mu_decreasing)



import pytest
from pinttr.exceptions import UnitsError

from eradiate.units import interpret_quantities, symbol
from eradiate.units import unit_context_config as ucc
from eradiate.units import unit_registry as ureg


def test_symbol():
    # Convert strings
    assert symbol("metre") == "m"
    assert symbol("dimensionless") == ""
    assert symbol("") == ""

    # Convert units
    assert symbol(ureg.dimensionless) == ""
    assert symbol(ureg.m) == "m"


def test_interpret_quantities():
    # If we pass no quantity map, behaviour is the same as pinttr.interpret_units()
    assert interpret_quantities({"wmin": 550.0}, {}, ucc) == {"wmin": 550.0}
    assert interpret_quantities({"wmin": 550.0, "wmin_units": "nm"}, {}, ucc) == {
        "wmin": 550.0 * ureg.nm
    }

    # If we pass a quantity map, unitless fields are applied units
    assert interpret_quantities({"wmin": 550.0}, {"wmin": "wavelength"}, ucc) == {
        "wmin": 550.0 * ureg.nm
    }

    # Unit context override propagates to default units
    with ucc.override(wavelength="micron"):
        assert interpret_quantities({"wmin": 550.0}, {"wmin": "wavelength"}, ucc) == {
            "wmin": 550.0 * ureg.micron
        }

    # Incompatible units raise
    with pytest.raises(UnitsError):
        assert interpret_quantities(
            {"wmin": 550.0, "wmin_units": "s"}, {"wmin": "wavelength"}, ucc
        )

    # Existing units are not modified...
    assert (
        interpret_quantities({"wmin": 0.55 * ureg.micron}, {"wmin": "wavelength"}, ucc)[
            "wmin"
        ].magnitude
        == 0.55
    )

    # ... unless force mode is used
    assert (
        interpret_quantities(
            {"wmin": 0.55 * ureg.micron}, {"wmin": "wavelength"}, ucc, force=True
        )["wmin"].magnitude
        == 550.0
    )



import numpy as np
import pytest

from eradiate.scenes.atmosphere import (
    ArrayParticleDistribution,
    ExponentialParticleDistribution,
    GaussianParticleDistribution,
    UniformParticleDistribution,
)


def test_particle_dist_uniform_construct():
    # Constructing without argument is possible
    UniformParticleDistribution()

    # 'bounds' must be sorted
    with pytest.raises(ValueError):
        UniformParticleDistribution([1.0, 0.0])

    # 'bounds' must be a 2-array
    with pytest.raises(ValueError):
        UniformParticleDistribution([0.0, 0.5, 1.0])


def test_particle_dist_uniform():
    dist = UniformParticleDistribution(bounds=[0.0, 0.5])
    x = np.linspace(-0.5, 1.0, 7)
    expected = np.array([0, 0, 2, 2, 2, 0, 0])
    assert np.allclose(dist(x), expected)


@pytest.mark.parametrize(
    "rate, scale, raises",
    [
        (None, None, False),
        (5.0, None, False),
        (None, 0.2, False),
        (1.0, 1.0, True),
    ],
)
def test_particle_dist_exponential(rate, scale, raises):
    if raises:
        with pytest.raises(ValueError):
            ExponentialParticleDistribution(rate=rate, scale=scale)
    else:
        dist = ExponentialParticleDistribution(rate=rate, scale=scale)
        x = np.linspace(0, 1, 11)
        expected = 5.0 * np.exp(-5.0 * x) / (1.0 - np.exp(-5.0))
        assert np.allclose(dist(x), expected)


def test_particle_dist_gaussian():
    dist = GaussianParticleDistribution(mean=0.0, std=1.0)
    x = np.linspace(0, 1, 11)
    expected = np.exp(-0.5 * np.square(x)) / np.sqrt(2.0 * np.pi)
    assert np.allclose(dist(x), expected)


def test_particle_dist_array_construct():
    # 'values' field is required
    with pytest.raises(TypeError):
        ArrayParticleDistribution()

    # Omitting the 'coords' field sets a default, regularly spaced grid
    dist = ArrayParticleDistribution(values=[0.1, 0.2, 0.3, 0.2, 0.1])
    assert np.allclose([0.1, 0.3, 0.5, 0.7, 0.9], dist.coords)

    # Specifying the 'coords' field requires both 'values' and 'coords' to be
    # of the same length
    with pytest.raises(ValueError):
        ArrayParticleDistribution(
            values=[0.1, 0.2, 0.3, 0.2, 0.1], coords=np.linspace(0, 1, 6)
        )


@pytest.mark.parametrize(
    "method, expected",
    [
        ("nearest", [0.1, 0.1, 0.1, 0.2, 0.2, 0.3, 0.3, 0.2, 0.2, 0.1, 0.1]),
        ("nearest-up", [0.1, 0.1, 0.2, 0.2, 0.2, 0.3, 0.2, 0.2, 0.1, 0.1, 0.1]),
        ("linear", [0.1, 0.1, 0.15, 0.2, 0.25, 0.3, 0.25, 0.2, 0.15, 0.1, 0.1]),
        ("zero", [0.1, 0.1, 0.1, 0.2, 0.2, 0.2, 0.3, 0.2, 0.2, 0.2, 0.1]),
        ("slinear", [0.1, 0.1, 0.15, 0.2, 0.25, 0.3, 0.25, 0.2, 0.15, 0.1, 0.1]),
        ("quadratic", [0.1, 0.1, 0.145, 0.2, 0.265, 0.3, 0.265, 0.2, 0.145, 0.1, 0.1]),
        (
            "cubic",
            [0.1, 0.1, 0.13125, 0.2, 0.26875, 0.3, 0.26875, 0.2, 0.13125, 0.1, 0.1],
        ),
        ("previous", [0.1, 0.1, 0.1, 0.2, 0.2, 0.2, 0.3, 0.2, 0.2, 0.2, 0.1]),
        ("next", [0.1, 0.1, 0.2, 0.2, 0.3, 0.3, 0.2, 0.2, 0.1, 0.1, 0.1]),
    ],
    ids=[
        "nearest",
        "nearest-up",
        "linear",
        "zero",
        "slinear",
        "quadratic",
        "cubic",
        "previous",
        "next",
    ],
)
def test_particle_dist_array_call(method, expected):
    dist = ArrayParticleDistribution(
        values=[0.1, 0.2, 0.3, 0.2, 0.1],
        method=method,
        extrapolate="nearest",
    )
    x = np.linspace(0, 1, 11)
    result = dist(x)
    assert np.allclose(result, expected)


@pytest.mark.parametrize(
    "extrapolate, expected",
    [
        ("zero", [0, 0]),
        ("nearest", [0.1, 0.1]),
        ("method", [0.05, 0.05]),
        ("nan", None),
    ],
    ids=["zero", "nearest", "method", "nan"],
)
def test_particle_dist_array_extrapolate(extrapolate, expected):
    dist = ArrayParticleDistribution(
        values=[0.1, 0.2, 0.3, 0.2, 0.1],
        method="linear",
        extrapolate=extrapolate,
    )
    result = dist([0.0, 1.0])

    if extrapolate != "nan":
        assert np.allclose(result, expected)
    else:
        assert np.isnan(result).all()



import pytest

import eradiate
from eradiate.experiments._helpers import measure_inside_atmosphere, surface_converter
from eradiate.scenes.atmosphere import HeterogeneousAtmosphere
from eradiate.scenes.surface import BasicSurface


def test_helpers_measure_inside_atmosphere(
    mode_mono,
    atmosphere_us_standard_mono,
):
    atm = HeterogeneousAtmosphere(
        geometry="plane_parallel",
        molecular_atmosphere=atmosphere_us_standard_mono,
    )

    s1 = eradiate.scenes.measure.MultiDistantMeasure()
    assert not measure_inside_atmosphere(atm, s1)

    s2 = eradiate.scenes.measure.PerspectiveCameraMeasure(
        origin=[1, 1, 1], target=[0, 0, 0]
    )
    assert measure_inside_atmosphere(atm, s2)

    s3 = eradiate.scenes.measure.MultiRadiancemeterMeasure(
        origins=[[1, 1, 1], [0, 0, 1000000]], directions=[[0, 0, -1], [0, 0, -1]]
    )
    with pytest.raises(ValueError):
        measure_inside_atmosphere(atm, s3)


def test_helpers_surface_converter(mode_mono):
    # A dictionary specifying a surface is converted to a surface
    assert isinstance(surface_converter({"type": "basic"}), BasicSurface)

    # A dictionary specifying a BSDF is converted to a surface
    assert isinstance(surface_converter({"type": "lambertian"}), BasicSurface)



"""
Utility functions for the Eradiate dependency management system.
"""

from copy import deepcopy
from typing import Literal


def resolve_package_manager(
    layered_yml: dict, manager: Literal["pip", "conda"]
) -> dict:
    """
    When relevant, filter out irrelevant dependency specifications in "packages"
    sections of the ``layered.yml`` file depending on the selected Python package
    manager.
    """
    result = {}

    for key, section in layered_yml.items():
        result[key] = deepcopy(
            section
        )  # We are going to mutate this, so we work on a copy

        package_list = result[key].get("packages", [])
        result[key]["packages"] = [
            package[manager] if isinstance(package, dict) else package
            for package in package_list
        ]

    return result



import astropy
import attrs
import dateutil
import numpy as np
import pytest
from numpy.testing import assert_allclose

import eradiate
from eradiate import unit_registry as ureg
from eradiate.spectral.index import SpectralIndex


@pytest.mark.slow
@pytest.mark.parametrize(
    "measure",
    [
        {"type": "hdistant"},
        {
            "type": "mdistant",
            "construct": "hplane",
            "zeniths": np.arange(-75, 76, 5),
            "azimuth": 0.0,
            "srf": {
                "type": "multi_delta",
                "wavelengths": 550.0 * ureg.nm,
            },
        },
    ],
    ids=["hdistant", "mdistant"],
)
@pytest.mark.parametrize("scale", [1.0, 0.5])
@pytest.mark.parametrize("datetime", [None, "2000-01-01"])
def test_radiance_scaling(modes_all_double, measure, scale, datetime):
    """
    Radiance scaling
    ================

    This test checks if the recorded radiance scales as expected with the
    incoming irradiance, while the reflectance remains constant.

    Rationale
    ---------

    An experiment is set up in which the illumination uses an irradiance
    spectrum to which a scaling is applied.

    Expected behaviour
    ------------------

    The scaling is applied to the illumination; consequently, the recorded
    radiance is scaled by the same amount.
    """

    reflectance = 0.5
    irradiance_spectrum = eradiate.scenes.spectra.SolarIrradianceSpectrum(
        dataset="thuillier_2003"
    )

    reference_irradiance = irradiance_spectrum.eval(SpectralIndex.new())

    exp = eradiate.experiments.AtmosphereExperiment(
        surface={"type": "lambertian", "reflectance": reflectance},
        atmosphere=None,
        illumination={
            "type": "directional",
            "irradiance": attrs.evolve(
                irradiance_spectrum, scale=scale, datetime=datetime
            ),
        },
        measures=measure,
    )
    result = eradiate.run(exp)

    # The radiance is proportional to the scaling factor
    scale_datetime = (
        (
            float(
                astropy.units.au
                / astropy.coordinates.get_sun(
                    astropy.time.Time(dateutil.parser.parse(datetime))
                ).distance
            )
            ** 2
        )
        if datetime is not None
        else 1.0
    )
    expected_radiance = (
        reference_irradiance.m * scale * scale_datetime * reflectance / np.pi
    )
    assert_allclose(result.radiance.values, expected_radiance, rtol=0.01)

    # The BRF is independent of the scaling factor
    assert_allclose(result.brf.values, reflectance)



"""Basic CKD correctness check."""

import numpy as np

import eradiate
from eradiate import unit_registry as ureg
from eradiate.experiments import AtmosphereExperiment
from eradiate.scenes.bsdfs import LambertianBSDF
from eradiate.scenes.measure import MultiDistantMeasure


def test_ckd_basic(modes_all_ckd):
    """
    CKD correctness check
    =====================

    We check for correctness when using CKD modes with atmosphere-free scenes.
    This test is designed to check that the CKD infrastructure
    (preprocessing, spectral loop, postprocessing) works.

    Rationale
    ---------

    Create an ``AtmosphereExperiment`` instance with no atmosphere, a Lambertian
    surface with reflectance equal to 1 and a ``MultiDistantMeasure``:

    * Sensor: Distant measure covering a plane (25 angular points,
      10^6 sample per pixel) and targeting (0, 0, 0).
    * Illumination: Directional illumination with a zenith angle
      :math:`\\theta = 0.0°` with black body irradiance spectrum.
    * Surface: a square surface with a Lambertian BRDF with 100 % reflectance.
    * Integrator: path tracer.

    Expected behaviour
    ------------------

    Recorded BRF must be exactly equal to one for all defined spectral bins.
    """

    # Configure experiment, run and postprocess results
    exp = AtmosphereExperiment(
        atmosphere=None,
        surface=LambertianBSDF(reflectance=1.0),
        measures=[
            MultiDistantMeasure.hplane(
                zeniths=np.arange(-60, 61, 5) * ureg.deg,
                azimuth=0.0 * ureg.deg,
                srf={
                    "type": "multi_delta",
                    "wavelengths": [
                        550,
                        560,
                        570,
                        510,
                    ]
                    * ureg.nm,  # We specify bins in arbitrary order on purpose
                },
            )
        ],
    )
    results = eradiate.run(exp, spp=1000)
    results = np.squeeze(results.data_vars["brf"].values)

    # Reflectance is uniform, equal to 1
    np.testing.assert_allclose(
        results, 1.0, rtol=1e-3 if eradiate.mode().is_single_precision else 1e-6
    )



import mitsuba as mi

from eradiate.scenes.phase import RayleighPhaseFunction
from eradiate.test_tools.types import check_scene_element


def test_rayleigh(modes_all_double):
    phase = RayleighPhaseFunction()
    check_scene_element(phase, mi.PhaseFunction)



import sys
from os.path import basename

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from docutils import nodes, statemachine
from docutils.parsers.rst import Directive


class ExecDirective(Directive):
    """Execute the specified python code and insert the output into the document"""

    # From https://stackoverflow.com/questions/7250659/how-to-use-python-to-programmatically-generate-part-of-sphinx-documentation

    has_content = True

    def run(self):
        oldStdout, sys.stdout = sys.stdout, StringIO()

        tab_width = self.options.get(
            "tab-width", self.state.document.settings.tab_width
        )
        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1
        )

        try:
            exec("\n".join(self.content))
            text = sys.stdout.getvalue()
            lines = statemachine.string2lines(text, tab_width, convert_whitespace=True)
            self.state_machine.insert_input(lines, source)
            return []
        except Exception:
            return [
                nodes.error(
                    None,
                    nodes.paragraph(
                        text="Unable to execute python code at %s:%d:"
                        % (basename(source), self.lineno)
                    ),
                    nodes.paragraph(text=str(sys.exc_info()[1])),
                )
            ]
        finally:
            sys.stdout = oldStdout


def setup(app):
    app.add_directive("exec", ExecDirective)
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import os
import tempfile

import mitsuba as mi
import pytest

from eradiate.scenes.shapes import FileMeshShape
from eradiate.test_tools.types import check_scene_element


@pytest.fixture(scope="module")
def tempfile_ply():
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, "tempfile_mesh.ply")
        with open(filename, "w") as tf:
            tf.write("ply\n")
            tf.write("format ascii 1.0\n")
            tf.write("element vertex 3\n")
            tf.write("property float x\n")
            tf.write("property float y\n")
            tf.write("property float z\n")
            tf.write("element face 1\n")
            tf.write("property list uchar int vertex_index\n")
            tf.write("end_header\n")
            tf.write("1 0 0\n")
            tf.write("0 1 0\n")
            tf.write("0 0 1\n")
            tf.write("3 1 2 3\n")
        yield filename


@pytest.fixture(scope="module")
def tempfile_obj():
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, "tempfile_mesh.obj")
        with open(filename, "w") as tf:
            tf.write("v 1 0 0\n")
            tf.write("v 0 1 0\n")
            tf.write("v 0 0 1\n")
            tf.write("f 1 2 3\n")
        yield filename


@pytest.mark.parametrize(
    "file_type, tempfile",
    [
        ("ply", "tempfile_ply"),
        ("obj", "tempfile_obj"),
    ],
    ids=["ply", "obj"],
)
def test_filemesh_kernel_dict(modes_all_double, file_type, tempfile, request):
    mesh = FileMeshShape(filename=request.getfixturevalue(tempfile))
    check_scene_element(mesh, mi.Mesh)


def test_construct_file_illegal(modes_all_double):
    with pytest.raises(ValueError, match="mesh files only in PLY or OBJ format"):
        FileMeshShape(filename="path/to/file.wrong")


@pytest.mark.parametrize(
    "file_type, tempfile",
    [
        ("ply", "tempfile_ply"),
        ("obj", "tempfile_obj"),
    ],
    ids=["ply", "obj"],
)
def test_construct_filemesh_trafo(mode_mono, file_type, tempfile, request):
    assert FileMeshShape(
        filename=request.getfixturevalue(tempfile), to_world=mi.Transform4f.scale(3)
    )



import mitsuba as mi
import numpy as np
import pytest
import xarray as xr

from eradiate.scenes.bsdfs import MQDiffuseBSDF
from eradiate.test_tools.types import check_scene_element

ds = xr.Dataset(
    {
        "brdf": xr.DataArray(
            data=np.full((3, 3, 3), 1.0 * np.pi),
            coords=[
                ("cos_theta_o", np.linspace(0, 1, 3), {"units": ""}),
                (
                    "phi_d",
                    np.linspace(0, 2.0 * np.pi, 3, endpoint=False),
                    {"units": "rad"},
                ),
                ("cos_theta_i", np.linspace(0, 1, 3), {"units": ""}),
            ],
        )
    }
)


def test_mqdiffuse_construct(modes_all):
    # Constructing with a well-formed dataset succeeds
    assert MQDiffuseBSDF(data=ds)

    # Missing 'brdf' variable fails
    with pytest.raises(ValueError, match="missing required data variable"):
        MQDiffuseBSDF(data=ds.rename({"brdf": "bsdf"}))

    # Incorrect dimension fails
    with pytest.raises(ValueError, match="incorrect dimension list"):
        MQDiffuseBSDF(data=ds.rename({"phi_d": "phi_r"}))

    # Incorrect coordinate fails
    with pytest.raises(
        ValueError, match="incorrect coordinate values for field 'phi_d'"
    ):
        MQDiffuseBSDF(
            data=ds.assign_coords(
                {
                    "phi_d": (
                        "phi_d",
                        np.linspace(0.0, 2.0 * np.pi, len(ds.phi_d)),
                        {"units": "rad"},
                    )
                }
            )
        )

    # Missing units fails
    with pytest.raises(
        ValueError,
        match="input dataset coordinate variable 'phi_d' is missing 'units' "
        "metadata field",
    ):
        MQDiffuseBSDF(
            data=ds.assign_coords(
                {
                    "phi_d": (
                        "phi_d",
                        np.linspace(0.0, 2.0 * np.pi, len(ds.phi_d), endpoint=True),
                    )
                }
            )
        )


def test_kernel_dict(mode_mono):
    bsdf = MQDiffuseBSDF(data=ds)
    check_scene_element(bsdf, mi.BSDF)



import mitsuba as mi

from eradiate.scenes.phase import IsotropicPhaseFunction
from eradiate.test_tools.types import check_scene_element


def test_isotropic(modes_all_double):
    # Default constructor
    phase = IsotropicPhaseFunction()
    check_scene_element(phase, mi.PhaseFunction)



import pytest

from eradiate.util.deprecation import DeprecatedWarning, UnsupportedWarning, deprecated


@pytest.mark.parametrize(
    "deprecation_type, expected_warning",
    (("deprecated", DeprecatedWarning), ("unsupported", UnsupportedWarning)),
)
def test_deprecated(deprecation_type, expected_warning):
    if deprecation_type == "deprecated":
        wrapper = deprecated(deprecated_in="1.1", current_version="1.2")

    elif deprecation_type == "unsupported":
        wrapper = deprecated(
            deprecated_in="1.1", removed_in="1.2", current_version="1.3"
        )

    else:
        raise ValueError("unhandled case")

    @wrapper
    def function():
        pass

    @wrapper
    class Class:
        @wrapper
        def instancemethod(self):
            pass

        @classmethod
        @wrapper
        def classmethod(cls):
            pass

        @staticmethod
        @wrapper
        def staticmethod():
            pass

    with pytest.warns(expected_warning):
        function()

    with pytest.warns(expected_warning):
        obj = Class()

    with pytest.warns(expected_warning):
        obj.instancemethod()

    with pytest.warns(expected_warning):
        Class.classmethod()

    with pytest.warns(expected_warning):
        Class.staticmethod()



from __future__ import annotations

import itertools
from pathlib import Path
from textwrap import dedent

import attrs

import eradiate
from eradiate.plot import dashboard_particle_dataset
from eradiate.scenes.biosphere import (
    RAMIActualCanopies,
    RAMIHeterogeneousAbstractCanopies,
)

eradiate.plot.set_style()

# Auto-generation disclaimer text
HEADER = dedent(
    """
    ..
      This file was automatically generated by docs/generate_rst_data.py. The

          make docs-rst-data

      target automates this process.
    """
).strip()


@attrs.define
class ParticleRadpropsInfo:
    keyword: str
    fname: str
    description: str | None = attrs.field(default=None)
    aliases: list[str] = attrs.field(factory=list)


PARTICLE_RADPROPS = [
    ParticleRadpropsInfo(
        "govaerts_2021-continental-extrapolated",
        "spectra/particles/govaerts_2021-continental-extrapolated.nc",
    ),
    ParticleRadpropsInfo(
        "govaerts_2021-desert-extrapolated",
        "spectra/particles/govaerts_2021-desert-extrapolated.nc",
    ),
]

RAMI_SCENE_COMMENTS = {
    "HET51_WWO_TLS": (
        "This version of the Wytham Wood scene uses data from the updated v2 "
        "dataset."
    )
}


def write_if_modified(filename, content):
    filename.parent.mkdir(parents=True, exist_ok=True)

    try:
        with open(filename, "r") as f:
            existing = f.read()
    except OSError:
        existing = None

    if existing == content:
        print(f"Skipping unchanged '{filename.name}'")

    else:
        print(f"Generating '{filename.name}'")
        with open(filename, "w") as f:
            f.write(content)


def savefig(fig, filename: Path, **kwargs):
    filename.parent.mkdir(exist_ok=True, parents=True)
    fig.savefig(filename, **kwargs)


def generate_particle_radprops_visual(
    info: ParticleRadpropsInfo, outfile: Path, force=False
):
    # Create dashboards for a dataset
    if outfile.is_file() and not force:  # Skip if file exists
        return

    with eradiate.data.open_dataset(info.fname) as ds:
        print(f"Generating particle radiative property visual in '{outfile}'")
        fig, _ = dashboard_particle_dataset(ds)
        savefig(fig, outfile, dpi=150)


def generate_particle_radprops_summary():
    # Write a table with the list of particle radiative property datasets
    outdir_visuals = Path(__file__).parent.absolute() / "rst/data/fig"
    outfile_rst = (
        Path(__file__).parent.absolute() / "rst/data/generated/aerosols_particles.rst"
    )
    print(f"Generating particle radiative property index in '{outfile_rst}'")

    sections = [
        dedent(
            """
Aerosol / particle single-scattering radiative properties
=========================================================

A particle radiative single-scattering property dataset provides collision 
coefficients and scattering phase matrix data for a given particle type. 
Eradiate's built-in particle radiative property datasets are managed by the 
data store (see :ref:`sec-data-intro` for details).

Format
------

* **Format** ``xarray.Dataset`` (in-memory), NetCDF (storage)
* **Dimensions**

  * ``w``: radiation wavelength
  * ``mu``: scattering angle cosine
  * ``i``: scattering phase matrix row index
  * ``j``: scattering phase matrix column index

* **Coordinates** (all dimension coordinates; when relevant, ``units`` are 
  required and specified in the units metadata field)

  * ``w`` float [length]
  * ``mu`` float [dimensionless]
  * ``i``,  ``j`` int

* **Data variables** (when relevant, units are required and  specified in the 
  units metadata field)

  * ``sigma_t`` (``w``): volume extinction coefficient [length^-1]
  * ``albedo`` (``w``): single-scattering albedo [dimensionless]
  * ``phase`` (``w``, ``mu``, ``i``, ``j``): scattering phase matrix 
    [steradian^-1]

* **Conventions**

  * Phase matrix components use C-style indexing (from 0).

.. dropdown:: Full validation schema

   .. literalinclude:: /resources/data_schemas/particle_dataset_v1.yml
"""
        ).strip()
    ]

    lst = ["\n".join(["Dataset index", "-------------"])]

    for info in PARTICLE_RADPROPS:
        outfile_visual = outdir_visuals / f"{info.keyword}.png"

        generate_particle_radprops_visual(info, outfile_visual)
        title = f"``{info.keyword}``"
        item = "\n".join(
            [
                title,
                "^" * len(title),
                "",
                f"Filename: ``{info.fname}``",
                "",
                f"{info.description if info.description else '*No description available.*'}",
                "",
                f".. image:: ../fig/{info.keyword}.png",
            ]
        )
        lst.append(item)

    sections.append("\n\n".join(lst))
    result = "\n\n".join([HEADER] + sections) + "\n"
    write_if_modified(outfile_rst, result)


def generate_rami_scene_summary():
    outfile_rst = (
        Path(__file__).parent.absolute() / "rst/data/generated/rami_scenes.rst"
    )
    print(f"Generating RAMI scene index in '{outfile_rst}'")

    from eradiate.scenes.biosphere import (
        RAMIHomogeneousAbstractCanopies,
    )

    sections = [
        """
RAMI benchmark scenes
=====================

Eradiate ships a collection of canopy scenes built from the 
`RAMI-V scene list <https://rami-benchmark.jrc.ec.europa.eu/_www/phase_descr.php?strPhase=RAMI5>`_
These pre-configured scenes are available for download request via the 
datastore.\ [#sn2]_ 
Usage is documented in the 
:ref:`scene loader guide <sec-user_guide-canopy_scene_loader-rami_scenes>`.

.. [#sn2] We thank the `DART <https://dart.omp.eu/>`_ team for letting us use 
   their 3D model files to derive our scene models.
   
.. note::

   The renders in the index table below use material spectra interpolated from 
   the monochromatic optical properties provided in the RAMI scene 
   specifications. Colours are therefore not realistic.

""".strip()
    ]

    table = [
        """
.. list-table:: RAMI benchmark scene index
   :widths: 1 1 1 2
   :header-rows: 1
   
   * - RAMI ID
     - Description
     - Comments      
     - Render
""".strip()
    ]

    for x in itertools.chain(
        RAMIHomogeneousAbstractCanopies,
        RAMIHeterogeneousAbstractCanopies,
        RAMIActualCanopies,
    ):
        rami_id = x.value
        description = x.name.replace("_", " ").lower()
        comments = RAMI_SCENE_COMMENTS.get(rami_id, "")
        row = f"""
   * - {rami_id}
     - {description}
     - {comments}
     - .. image:: /fig/rami_scenes/{x.name}_30_90.png
"""
        table.append(row.strip("\n"))

    sections.append("\n".join(table))

    result = "\n\n".join([HEADER] + sections) + "\n"
    write_if_modified(outfile_rst, result)


if __name__ == "__main__":
    generate_particle_radprops_summary()
    generate_rami_scene_summary()



import pytest

from eradiate.data import data_store


@pytest.mark.parametrize(
    "path",
    [
        "chemistry/molecular_masses.nc",
        "spectra/solar_irradiance/blackbody_sun.nc",
        "spectra/solar_irradiance/thuillier_2003.nc",
        "spectra/solar_irradiance/meftah_2018.nc",
        "spectra/solar_irradiance/solid_2017-mean.nc",
        "spectra/solar_irradiance/whi_2008-faculae_active.nc",
        "thermoprops/afgl_1986-tropical.nc",
        "thermoprops/afgl_1986-midlatitude_summer.nc",
        "thermoprops/afgl_1986-midlatitude_winter.nc",
        "thermoprops/afgl_1986-subarctic_summer.nc",
        "thermoprops/afgl_1986-subarctic_winter.nc",
        "thermoprops/afgl_1986-us_standard.nc",
    ],
)
def test_required_data_sets(path):
    # Data sets required for testing are found
    assert data_store.fetch(path)
